"""Physical fields output (:mod:`fluidsim.base.output.phys_fields`)
===================================================================

Provides:

.. autoclass:: PhysFieldsBase
   :members:
   :private-members:

.. autoclass:: SetOfPhysFieldFiles
   :members:
   :private-members:

"""

import re
import os
from glob import glob
from pathlib import Path
from math import isclose

import numpy as np
import h5py

from fluiddyn.util import mpi

from fluidsim.util.output import save_file, h5pack, ext

from .base import SpecificOutput


class PhysFieldsBase(SpecificOutput):
    """Manage the output of physical fields."""

    _tag = "phys_fields"

    @staticmethod
    def _complete_params_with_default(params):
        tag = "phys_fields"
        params.output._set_child(
            tag, attribs={"field_to_plot": "ux", "file_with_it": False}
        )

        params.output.periods_save._set_attrib(tag, 0)
        params.output.periods_plot._set_attrib(tag, 0)

    def __init__(self, output):
        params = output.sim.params
        self.output = output
        self.oper = output.oper

        if hasattr(self, "_init_skip_quiver"):
            self._init_skip_quiver()

        if hasattr(self, "_init_movies"):
            self._init_movies()
            self.animate = self.movies.animate
            self.interact = self.movies.interact

        super().__init__(
            output,
            period_save=params.output.periods_save.phys_fields,
            period_plot=params.output.periods_plot.phys_fields,
        )

        self.key_field_to_plot = self.get_key_field_to_plot()

        self.set_of_phys_files = SetOfPhysFieldFiles(output=self.output)
        self._equation = None

        if self.period_save == 0 and self.period_plot == 0:
            self.t_last_save = -np.inf
            return

        self.t_last_save = self.sim.time_stepping.t
        self.t_last_plot = self.sim.time_stepping.t

    def get_key_field_to_plot(self, forbid_compute=False, key_field_to_plot=None):
        params = self.params
        if key_field_to_plot is None:
            key_field_to_plot = params.output.phys_fields.field_to_plot
        info_state = self.sim.info.solver.classes.State
        keys_ok = info_state.keys_state_phys.copy()

        if not params.ONLY_COARSE_OPER and not forbid_compute:
            keys_ok.extend(info_state.keys_computable)

        if key_field_to_plot not in keys_ok:
            key_field_to_plot = info_state.keys_state_phys[0]
            for key in ["q", "rot", "rotz"]:
                if key in keys_ok:
                    key_field_to_plot = key

        return key_field_to_plot

    def _init_path_files(self):
        super()._init_path_files()

        # This if clause is required since the function _init_path_files is
        # first called by the super().__init__ function when set_of_phys_files
        # is not initialized (see above). Useful when end_of_simul is called.
        if hasattr(self, "set_of_phys_files"):
            self.set_of_phys_files.path_dir = self.output.path_run

    def _init_files(self, arrays_1st_time=None):
        # Does nothing on purpose...
        pass

    def _init_online_plot(self):
        pass

    def _online_save(self):
        """Online save."""
        tsim = self.sim.time_stepping.t
        if self._has_to_online_save():
            self.t_last_save = tsim
            self.save()

    def _online_plot(self):
        """Online plot."""
        tsim = self.sim.time_stepping.t
        if tsim - self.t_last_plot >= self.period_plot:
            self.t_last_plot = tsim
            itsim = self.sim.time_stepping.it
            self.plot(
                numfig=itsim,
                key_field=self.params.output.phys_fields.field_to_plot,
            )

    def save(self, state_phys=None, params=None, particular_attr=None):
        if state_phys is None:
            state_phys = self.sim.state.state_phys
        if params is None:
            params = self.params

        time = self.sim.time_stepping.t

        path_run = Path(self.output.path_run)

        if params.time_stepping.USE_T_END:
            # check if some state_phys files already exist
            try:
                path_test = next(path_run.glob("state_phys*"))
            except StopIteration:
                # file does not exist : get str_width from t_end
                # max number of digits = int(log10(t_end)) + 1
                # add .3f precision = 4 additional characters
                # +2 by anticipation of potential restarts
                str_width = int(np.log10(params.time_stepping.t_end)) + 7
            else:
                # file does exist : get str_width from file name
                # file name is something like 'state_phys_tYYYY.YYY.nc'
                str_width = len(path_test.name[12:-3])
        else:
            # dynamic width not implemented if USE_T_END==False
            str_width = 7

        if mpi.rank == 0:
            path_run.mkdir(exist_ok=True)

        if 0 < self.period_save < 0.001 or params.output.phys_fields.file_with_it:
            str_it = f"_it={self.sim.time_stepping.it}"
        else:
            str_it = ""

        name_save = f"state_phys_t{time:0{str_width}.3f}{str_it}.{ext}"

        path_file = path_run / name_save

        does_path_exist = None
        if mpi.rank == 0:
            does_path_exist = path_file.exists()
        if mpi.nb_proc > 1:
            does_path_exist = mpi.comm.bcast(does_path_exist, root=0)

        if does_path_exist:
            # do not save if the file corresponds to the same it
            it_file = None
            if mpi.rank == 0:
                with h5pack.File(str(path_file), "r") as file:
                    it_file = file["state_phys"].attrs["it"]
            if mpi.nb_proc > 1:
                it_file = mpi.comm.bcast(it_file, root=0)
            if it_file == self.sim.time_stepping.it:
                return
            name_save = (
                f"state_phys_t{time:07.3f}_it={self.sim.time_stepping.it}.{ext}"
            )
            path_file = path_run / name_save
        self.output.print_stdout("save state_phys in file " + name_save)

        save_file(
            path_file,
            state_phys,
            self.sim.info,
            self.output.name_run,
            self.sim.oper,
            time,
            self.sim.time_stepping.it,
            particular_attr,
        )

    def get_field_to_plot(
        self,
        key=None,
        time=None,
        idx_time=None,
        equation=None,
        interpolate_time=True,
    ):
        """Get the field to be plotted in process 0."""

        if equation is None:
            equation = self._equation

        if (
            not self.sim.params.ONLY_COARSE_OPER
            and (time is None or isclose(time, self.sim.time_stepping.t))
            and idx_time is None
        ):
            # we get the field from the state
            field, key = self.get_field_to_plot_from_state(
                field=key, equation=equation
            )
            return field, self.sim.time_stepping.t

        else:
            return self.set_of_phys_files.get_field_to_plot(
                time=time,
                idx_time=idx_time,
                key=key,
                equation=equation,
                interpolate_time=interpolate_time,
            )

    def get_field_to_plot_from_state(self, field=None, equation=None):
        """Get the field to be plotted in process 0."""

        if field is None:
            key_field = self.key_field_to_plot
            field_loc = self.sim.state.get_var(key_field)
        elif isinstance(field, np.ndarray):
            key_field = "given field"
            field_loc = field
        else:
            field_loc = self.sim.state.get_var(field)
            key_field = field

        if mpi.nb_proc > 1:
            field = self.oper.gather_Xspace(field_loc)
        else:
            field = field_loc

        if equation is None:
            return field, key_field

        elif equation.startswith("iz="):
            iz = eval(equation[len("iz=") :])
            field = field[iz, ...]
        elif equation.startswith("z="):
            z = eval(equation[len("z=") :])
            iz = abs(self.output.sim.oper.get_grid1d_seq("z") - z).argmin()
            field = field[iz, ...]
        elif equation.startswith("iy="):
            iy = eval(equation[len("iy=") :])
            field = field[:, iy, :]
        elif equation.startswith("y="):
            y = eval(equation[len("y=") :])
            iy = abs(self.output.sim.oper.get_grid1d_seq("y") - y).argmin()
            field = field[:, iy, :]
        elif equation.startswith("ix="):
            ix = eval(equation[len("ix=") :])
            field = field[..., ix]
        elif equation.startswith("x="):
            x = eval(equation[len("x=") :])
            ix = abs(self.output.sim.oper.get_grid1d_seq("x") - x).argmin()
            field = field[..., ix]
        else:
            raise NotImplementedError

        return field, key_field


def time_from_path(path):
    """Regular expression search to extract time from filename."""
    filename = os.path.basename(path)
    pattern = r"""
        (?!t)     # text after t but exclude it
        [0-9]+    # a couple of digits
        \.        # the decimal point
        [0-9]+    # a couple of digits
    """
    match = re.search(pattern, filename, re.VERBOSE)
    time = float(match.group(0))
    return time


class SetOfPhysFieldFiles:
    """A set of physical field files."""

    def __init__(self, path_dir=os.curdir, output=None):
        self.output = output
        self.path_dir = path_dir if output is None else output.path_run
        self.update_times()

    def update_times(self):
        """Initialize the times by globing and analyzing the file names."""
        path_files = glob(os.path.join(self.path_dir, "state_phys*.[hn]*"))

        if hasattr(self, "path_files") and len(self.path_files) == len(
            path_files
        ):
            return

        self.path_files = sorted(path_files)
        self.times = np.array([time_from_path(path) for path in self.path_files])

    def get_min_time(self):
        if hasattr(self, "times"):
            return self.times.min()
        return 0.

    def get_field_to_plot(
        self,
        time=None,
        idx_time=None,
        key=None,
        equation=None,
        interpolate_time=True,
    ):

        if time is None and idx_time is None:
            self.update_times()
            time = self.times[-1]

        # Assert files are available
        if self.times.size == 0:
            raise FileNotFoundError(
                "No state_phys files were detected in directory: "
                f"{self.path_dir}"
            )

        if not interpolate_time and time is not None:
            idx, time_closest = self.get_closest_time_file(time)
            return self.get_field_to_plot(
                idx_time=idx, key=key, equation=equation
            )

        if interpolate_time and time is not None:

            idx_closest, time_closest = self.get_closest_time_file(time)

            if isclose(time, time_closest) or self.times.size == 1:
                return self.get_field_to_plot(
                    idx_time=idx_closest, key=key, equation=equation
                )

            if idx_closest == self.times.size - 1:
                idx0 = idx_closest - 1
                idx1 = idx_closest
            elif time_closest < time or idx_closest == 0:
                idx0 = idx_closest
                idx1 = idx_closest + 1
            elif time_closest > time:
                idx0 = idx_closest - 1
                idx1 = idx_closest

            dt_save = self.times[idx1] - self.times[idx0]
            weight0 = 1 - np.abs(time - self.times[idx0]) / dt_save
            weight1 = 1 - np.abs(time - self.times[idx1]) / dt_save

            field0, time0 = self.get_field_to_plot(
                idx_time=idx0, key=key, equation=equation
            )
            field1, time1 = self.get_field_to_plot(
                idx_time=idx1, key=key, equation=equation
            )

            return field0 * weight0 + field1 * weight1, time

        # print(idx_time, 'Using file', self.path_files[idx_time])

        with h5py.File(self.path_files[idx_time], "r") as file:
            time = file["state_phys"].attrs["time"]
            dset = file["state_phys"][key]

            if equation is None:
                return dset[...], time

            if equation.startswith("iz="):
                iz = eval(equation[len("iz=") :])
                return dset[iz, ...], time

            elif equation.startswith("z="):
                z = eval(equation[len("z=") :])
                iz = abs(self.output.sim.oper.get_grid1d_seq("z") - z).argmin()
                return dset[iz, ...], time

            elif equation.startswith("iy="):
                iy = eval(equation[len("iy=") :])
                return dset[:, iy, :], time

            elif equation.startswith("y="):
                y = eval(equation[len("y=") :])
                iy = abs(self.output.sim.oper.get_grid1d_seq("y") - y).argmin()
                return dset[:, iy, :], time

            elif equation.startswith("ix="):
                ix = eval(equation[len("ix=") :])
                return dset[..., ix], time

            elif equation.startswith("x="):
                x = eval(equation[len("x=") :])
                ix = abs(self.output.sim.oper.get_grid1d_seq("x") - x).argmin()
                return dset[..., ix], time

            else:
                raise NotImplementedError

    def get_closest_time_file(self, time):
        """Find the index and value of the closest actual time of the field."""
        idx = np.abs(self.times - time).argmin()
        return idx, self.times[idx]



"""SEB (:mod:`fluidsim.solvers.ns3d.strat.output.spect_energy_budget`)
======================================================================

.. autoclass:: SpectralEnergyBudgetNS3DStrat
   :members:
   :private-members:

"""

import numpy as np

from warnings import warn
from fluidsim.solvers.ns3d.output.spect_energy_budget import (
    SpectralEnergyBudgetNS3D,
)


class SpectralEnergyBudgetNS3DStrat(SpectralEnergyBudgetNS3D):
    r"""Spectral energy budget of ns3d.strat.

    .. |vv| mathmacro:: \textbf{v}

    .. |kk| mathmacro:: \textbf{k}

    .. |convK2A| mathmacro:: C_{K\rightarrow A}

    .. |bnabla| mathmacro:: \boldsymbol{\nabla}

    .. |bomega| mathmacro:: \boldsymbol{\omega}

    Notes
    -----

    .. math::

      d_t E_K(\kk) = T_K(\kk) - \convK2A(\kk) - D_K(\kk),

      d_t E_A(\kk) = T_A(\kk) + \convK2A(\kk) - D_A(\kk),

    where :math:`E_K(\kk) = |\hat{\vv}|^2/2` and :math:`E_A(\kk) = |\hat{b}|^2/(2N^2)`.

    The transfer terms are

    .. math::

      T_K(\kk) = \Re \left(\hat{\vv}^* \cdot P_\perp\widehat{\vv \times \bomega} \right),

      T_A(\kk) = - \Re \left(\hat{b}^* \widehat{\vv \cdot \bnabla b} \right) / N^2.

    By definition, we have to have :math:`\sum T_K(\kk) = \sum T_A(\kk) = 0`.

    The conversion term is equal to :math:`\convK2A(\kk) = -\Re(\hat{v_z}^*
    \hat{b})` and the dissipative terms are

    .. math::

      D_K(\kk) = f_{dK}(\kk) E_K(\kk),

      D_A(\kk) = f_{dA}(\kk) E_A(\kk),

    where :math:`f_{dK}(\kk)` and :math:`f_{dA}(\kk)` are the dissipation
    frequency depending of the wavenumber and the viscous coefficients.

    """

    def compute(self):

        results = super().compute()

        state = self.sim.state
        N = self.sim.params.N

        oper = self.sim.oper

        state_spect = state.state_spect

        vz_fft = state_spect.get_var("vz_fft")

        state_phys = state.state_phys

        f_d, f_d_hypo = self.sim.compute_freq_diss()
        del f_d_hypo

        b_fft = state_spect.get_var("b_fft")

        results.update(
            self.compute_spectra("diss_A", 1 / N ** 2 * f_d * abs(b_fft) ** 2)
        )

        results.update(
            self.compute_spectra("conv_K2A", -np.real(vz_fft.conj() * b_fft))
        )
        del vz_fft

        vx = state_phys.get_var("vx")
        vy = state_phys.get_var("vy")
        vz = state_phys.get_var("vz")

        fb_fft = (
            -1
            / N ** 2
            * oper.div_vb_fft_from_vb(vx, vy, vz, state_phys.get_var("b"))
        )
        del vx, vy, vz

        results.update(
            self.compute_spectra("transfer_A", np.real(b_fft.conj() * fb_fft))
        )

        for key, value in results.items():
            if key.startswith("transfer_A"):
                try:
                    assert value.sum() < 1e-12
                except AssertionError:
                    warn(
                        f"spect_energy_budg: transfer_A.sum() is too big {value.sum()}"
                    )

        return results

    def plot_fluxes(
        self, tmin=None, tmax=None, key_k="kh", ax=None, plot_conversion=True
    ):

        data = self.compute_fluxes_mean(tmin, tmax)

        k_plot = data[key_k]
        deltak = k_plot[1]
        k_plot += deltak / 2

        key_flux = key_k[1] + "flux_"
        flux_Kh = data[key_flux + "transfer_Kh"]
        flux_Kz = data[key_flux + "transfer_Kz"]
        flux_A = data[key_flux + "transfer_A"]
        DKh = data[key_flux + "diss_Kh"]
        DKz = data[key_flux + "diss_Kz"]
        DA = data[key_flux + "diss_A"]
        if plot_conversion:
            CK2A = data[key_flux + "conv_K2A"]

        flux_K = flux_Kh + flux_Kz
        flux_tot = flux_K + flux_A
        DK = DKh + DKz
        D = DK + DA

        eps = D[-1]

        if ax is None:
            fig, ax = self.output.figure_axe()

        xlabel = "k_" + key_k[1]
        ylabel = fr"$\Pi({xlabel})/\epsilon$"
        ax.set_xlabel(f"${xlabel}$")
        ax.set_ylabel(ylabel)
        ax.set_title(f"spectral fluxes\n{self.output.summary_simul}")

        def _plot(x, y, code, label, linewidth=None):
            ax.semilogx(x, y / eps, code, label=label, linewidth=linewidth)

        _plot(k_plot, flux_tot, "k", r"$\Pi/\epsilon$", linewidth=2)
        _plot(k_plot, D, "k--", r"$D/\epsilon$", linewidth=2)
        _plot(k_plot, flux_tot + D, "k:", r"$(\Pi+D)/\epsilon$")
        _plot(k_plot, flux_K, "r", r"$\Pi_K/\epsilon$")
        _plot(k_plot, flux_A, "b", r"$\Pi_A/\epsilon$")
        _plot(k_plot, DK, "r--", r"$D_K/\epsilon$")
        _plot(k_plot, DA, "b--", r"$D_A/\epsilon$")
        if plot_conversion:
            _plot(k_plot, CK2A, "m", r"$B/\epsilon$", linewidth=2)

        ax.legend()



"""Operators sw1l (:mod:`fluidsim.operators.operators`)
=======================================================

Provides

.. autoclass:: OperatorsPseudoSpectralSW1L
   :members:
   :private-members:

"""

import numpy as np
from fluiddyn.util.compat import cached_property

from transonic import boost, jit, Array

from fluidsim.operators.operators2d import (
    OperatorsPseudoSpectral2D,
    rank,
    laplacian_fft,
    invlaplacian_fft,
)

AC = Array[np.complex128, "2d"]
AF = Array[np.float64, "2d"]


@jit
def _qapamfft_from_uxuyetafft(
    ux_fft: AC,
    uy_fft: AC,
    eta_fft: AC,
    n0: int,
    n1: int,
    KX: AF,
    KY: AF,
    K2: AF,
    Kappa_over_ic: AC,
    f: float,
    c2: float,
    rank: int,
):
    """Calculate normal modes from primitive variables."""
    freq_Corio = f
    f_over_c2 = freq_Corio / c2

    q_fft = np.empty([n0, n1], dtype=np.complex128)
    ap_fft = np.empty([n0, n1], dtype=np.complex128)
    am_fft = np.empty([n0, n1], dtype=np.complex128)

    if freq_Corio != 0:
        for i0 in range(n0):
            for i1 in range(n1):
                if i0 == 0 and i1 == 0 and rank == 0:
                    q_fft[i0, i1] = 0
                    ap_fft[i0, i1] = ux_fft[0, 0] + 1.0j * uy_fft[0, 0]
                    am_fft[i0, i1] = ux_fft[0, 0] - 1.0j * uy_fft[0, 0]
                else:

                    rot_fft = 1j * (
                        KX[i0, i1] * uy_fft[i0, i1] - KY[i0, i1] * ux_fft[i0, i1]
                    )

                    q_fft[i0, i1] = rot_fft - freq_Corio * eta_fft[i0, i1]

                    a_over2_fft = 0.5 * (
                        K2[i0, i1] * eta_fft[i0, i1] + f_over_c2 * rot_fft
                    )

                    Deltaa_over2_fft = (
                        0.5j
                        * Kappa_over_ic[i0, i1]
                        * (
                            KX[i0, i1] * ux_fft[i0, i1]
                            + KY[i0, i1] * uy_fft[i0, i1]
                        )
                    )

                    ap_fft[i0, i1] = a_over2_fft + Deltaa_over2_fft
                    am_fft[i0, i1] = a_over2_fft - Deltaa_over2_fft

    else:  # (freq_Corio == 0.)
        for i0 in range(n0):
            for i1 in range(n1):
                if i0 == 0 and i1 == 0 and rank == 0:
                    q_fft[i0, i1] = 0
                    ap_fft[i0, i1] = ux_fft[0, 0] + 1.0j * uy_fft[0, 0]
                    am_fft[i0, i1] = ux_fft[0, 0] - 1.0j * uy_fft[0, 0]
                else:
                    q_fft[i0, i1] = 1j * (
                        KX[i0, i1] * uy_fft[i0, i1] - KY[i0, i1] * ux_fft[i0, i1]
                    )

                    a_over2_fft = 0.5 * K2[i0, i1] * eta_fft[i0, i1]

                    Deltaa_over2_fft = (
                        0.5j
                        * Kappa_over_ic[i0, i1]
                        * (
                            KX[i0, i1] * ux_fft[i0, i1]
                            + KY[i0, i1] * uy_fft[i0, i1]
                        )
                    )

                    ap_fft[i0, i1] = a_over2_fft + Deltaa_over2_fft
                    am_fft[i0, i1] = a_over2_fft - Deltaa_over2_fft

    return q_fft, ap_fft, am_fft


@boost
class OperatorsPseudoSpectralSW1L(OperatorsPseudoSpectral2D):
    Kappa_over_ic: AC
    nK0_loc: int
    nK1_loc: int
    rank: int

    @cached_property
    def Kappa2_not0(self):
        return self.K2_not0 + self.params.kd2

    @cached_property
    def Kappa_over_ic(self):
        Kappa2 = self.K2 + self.params.kd2
        return -1.0j * np.sqrt(Kappa2 / self.params.c2)

    def qapamfft_from_uxuyetafft(self, ux_fft, uy_fft, eta_fft, params=None):
        """ux, uy, eta (fft) ---> q, ap, am (fft)"""

        if params is None:
            params = self.params

        n0 = self.nK0_loc
        n1 = self.nK1_loc

        KX = self.KX
        KY = self.KY
        K2 = self.K2
        Kappa_over_ic = self.Kappa_over_ic
        f = float(params.f)
        c2 = float(params.c2)

        return _qapamfft_from_uxuyetafft(
            ux_fft,
            uy_fft,
            eta_fft,
            n0,
            n1,
            KX,
            KY,
            K2,
            Kappa_over_ic,
            f,
            c2,
            rank,
        )

    def uxuyetafft_from_qapamfft(self, q_fft, ap_fft, am_fft):
        """q, ap, am (fft) ---> ux, uy, eta (fft)"""
        a_fft = ap_fft + am_fft
        if rank == 0:
            a_fft[0, 0] = 0.0
        div_fft = self.divfft_from_apamfft(ap_fft, am_fft)
        (uxa_fft, uya_fft, etaa_fft) = self.uxuyetafft_from_afft(a_fft)
        (uxq_fft, uyq_fft, etaq_fft) = self.uxuyetafft_from_qfft(q_fft)
        uxd_fft, uyd_fft = self.vecfft_from_divfft(div_fft)
        ux_fft = uxa_fft + uxq_fft + uxd_fft
        uy_fft = uya_fft + uyq_fft + uyd_fft
        eta_fft = etaa_fft + etaq_fft
        if rank == 0:
            ux_fft[0, 0] = 0.5 * (ap_fft[0, 0] + am_fft[0, 0])
            uy_fft[0, 0] = 0.5j * (am_fft[0, 0] - ap_fft[0, 0])
        return ux_fft, uy_fft, eta_fft

    def vecfft_from_rotdivfft(self, rot_fft, div_fft):
        """Inverse of the Helmholtz decomposition."""
        # TODO: Pythranize
        urx_fft, ury_fft = self.vecfft_from_rotfft(rot_fft)
        udx_fft, udy_fft = self.vecfft_from_divfft(div_fft)
        ux_fft = urx_fft + udx_fft
        uy_fft = ury_fft + udy_fft
        return ux_fft, uy_fft

    def uxuyetafft_from_qfft(self, q_fft, params=None):
        """Compute ux, uy and eta in Fourier space."""
        if params is None:
            params = self.params
            Kappa2_not0 = self.Kappa2_not0
        else:
            Kappa2_not0 = self.K2_not0 + params.kd2

        ilq_fft = invlaplacian_fft(q_fft, Kappa2_not0, rank)
        rot_fft = laplacian_fft(ilq_fft, self.K2)
        ux_fft, uy_fft = self.vecfft_from_rotfft(rot_fft)

        if params.f == 0:
            eta_fft = self.create_arrayK(value=0)
        else:
            eta_fft = -params.f * ilq_fft / params.c2
        return ux_fft, uy_fft, eta_fft

    def uxuyetafft_from_afft(self, a_fft, params=None):
        """Compute ux, uy and eta in Fourier space."""
        if params is None:
            params = self.params

        eta_fft = self.etafft_from_afft(a_fft, params)
        if params.f == 0:
            rot_fft = self.create_arrayK(value=0)
        else:
            rot_fft = params.f * eta_fft

        ux_fft, uy_fft = self.vecfft_from_rotfft(rot_fft)

        return ux_fft, uy_fft, eta_fft

    def rotfft_from_qfft(self, q_fft, params=None):
        """Compute ux, uy and eta in Fourier space."""
        if params is None:
            params = self.params
            Kappa2_not0 = self.Kappa2_not0
        else:
            Kappa2_not0 = self.K2_not0 + params.kd2

        rot_fft = laplacian_fft(
            invlaplacian_fft(q_fft, Kappa2_not0, rank), self.K2
        )
        return rot_fft

    def rotfft_from_afft(self, a_fft, params=None):
        """Compute ux, uy and eta in Fourier space."""
        if params is None:
            params = self.params

        if params.f == 0:
            rot_fft = self.create_arrayK(value=0)
        else:
            rot_fft = params.f * self.etafft_from_afft(a_fft, params)

        return rot_fft

    def afft_from_uxuyetafft(self, ux_fft, uy_fft, eta_fft, params=None):
        if params is None:
            params = self.params

        a_fft = laplacian_fft(eta_fft, self.K2)
        if params.f != 0:
            rot_fft = self.rotfft_from_vecfft(ux_fft, uy_fft)
            a_fft += params.f / params.c2 * rot_fft
        return a_fft

    def etafft_from_qfft(self, q_fft, params=None):
        """Compute eta in Fourier space."""
        if params is None:
            params = self.params
            Kappa2_not0 = self.Kappa2_not0
        else:
            Kappa2_not0 = self.K2_not0 + params.kd2

        if params.f == 0:
            eta_fft = self.create_arrayK(value=0)
        else:
            eta_fft = (
                -params.f / params.c2 * invlaplacian_fft(q_fft, Kappa2_not0, rank)
            )
        return eta_fft

    def etafft_from_afft(self, a_fft, params=None):
        """Compute eta in Fourier space."""
        if params is None:
            params = self.params
            Kappa2_not0 = self.Kappa2_not0
        else:
            Kappa2_not0 = self.K2_not0 + params.kd2

        eta_fft = invlaplacian_fft(a_fft, Kappa2_not0, rank)
        return eta_fft

    def etafft_from_aqfft(self, a_fft, q_fft, params=None):
        """Compute eta in Fourier space."""
        if params is None:
            params = self.params
            Kappa2_not0 = self.Kappa2_not0
        else:
            Kappa2_not0 = self.K2_not0 + params.kd2

        if params.f == 0:
            eta_fft = invlaplacian_fft(a_fft, self.K2_not0, rank)
        else:
            eta_fft = invlaplacian_fft(
                (a_fft - params.f / params.c2 * q_fft), Kappa2_not0, rank
            )
        return eta_fft

    def qdafft_from_uxuyetafft(self, ux_fft, uy_fft, eta_fft, params=None):
        if params is None:
            params = self.params
        div_fft = self.divfft_from_vecfft(ux_fft, uy_fft)
        rot_fft = self.rotfft_from_vecfft(ux_fft, uy_fft)
        q_fft = rot_fft - params.f * eta_fft
        ageo_fft = params.f / params.c2 * rot_fft + laplacian_fft(
            eta_fft, self.K2
        )
        return q_fft, div_fft, ageo_fft

    def apamfft_from_adfft(self, a_fft, d_fft):
        """Return the eigen modes ap and am."""
        Delta_a_fft = self.Kappa_over_ic * d_fft
        ap_fft = 0.5 * (a_fft + Delta_a_fft)
        am_fft = 0.5 * (a_fft - Delta_a_fft)
        return ap_fft, am_fft

    @jit
    def divfft_from_apamfft(self, ap_fft: AC, am_fft: AC):
        """Return div from the eigen modes ap and am."""
        n0 = self.nK0_loc
        n1 = self.nK1_loc
        Kappa_over_ic = self.Kappa_over_ic
        rank = self.rank

        Delta_a_fft = ap_fft - am_fft
        d_fft = np.empty([n0, n1], dtype=np.complex128)

        for i0 in range(n0):
            for i1 in range(n1):
                if i0 == 0 and i1 == 0 and rank == 0:
                    d_fft[i0, i1] = 0.0
                else:
                    d_fft[i0, i1] = Delta_a_fft[i0, i1] / Kappa_over_ic[i0, i1]
        return d_fft

    def qapamfft_from_etafft(self, eta_fft, params=None):
        """eta (fft) ---> q, ap, am (fft)"""
        if params is None:
            params = self.params

        q_fft = -params.f * eta_fft
        ap_fft = 0.5 * laplacian_fft(eta_fft, self.K2)
        am_fft = ap_fft.copy()
        return q_fft, ap_fft, am_fft



"""Output for the ns3d.strat solver
===================================

.. autoclass:: Output
   :members:
   :private-members:

.. autosummary::
   :toctree:

   spatial_means
   spectra
   spect_energy_budget

"""

try:
    from fluidsim.operators.operators3d import (
        compute_energy_from_1field,
        compute_energy_from_1field_with_coef,
        compute_energy_from_2fields,
        compute_energy_from_3fields,
    )
except ImportError:
    from fluidsim import _is_testing

    if not _is_testing:
        raise

from ...output import Output as OutputNS3D


class Output(OutputNS3D):
    """Main output class for ns3d.strat"""

    @staticmethod
    def _complete_info_solver(info_solver):
        """Complete the ParamContainer info_solver."""
        OutputNS3D._complete_info_solver(info_solver)

        classes = info_solver.classes.Output.classes
        base_name_mod = "fluidsim.solvers.ns3d.strat.output"

        classes.SpatialMeans.module_name = base_name_mod + ".spatial_means"
        classes.SpatialMeans.class_name = "SpatialMeansNS3DStrat"

        classes.Spectra.module_name = base_name_mod + ".spectra"
        classes.Spectra.class_name = "SpectraNS3DStrat"

        classes.SpectralEnergyBudget.module_name = (
            base_name_mod + ".spect_energy_budget"
        )
        classes.SpectralEnergyBudget.class_name = "SpectralEnergyBudgetNS3DStrat"

    def compute_energies_fft(self):
        get_var = self.sim.state.state_spect.get_var
        b_fft = get_var("b_fft")
        vx_fft = get_var("vx_fft")
        vy_fft = get_var("vy_fft")
        vz_fft = get_var("vz_fft")

        urx_fft, ury_fft, udx_fft, udy_fft = self.oper.urudfft_from_vxvyfft(
            vx_fft, vy_fft
        )

        nrj_A = compute_energy_from_1field_with_coef(
            b_fft, 1.0 / self.sim.params.N ** 2
        )
        nrj_Kz = compute_energy_from_1field(vz_fft)
        nrj_Khr = compute_energy_from_2fields(urx_fft, ury_fft)
        nrj_Khd = compute_energy_from_2fields(udx_fft, udy_fft)
        return nrj_A, nrj_Kz, nrj_Khr, nrj_Khd

    def compute_energy_fft(self):
        get_var = self.sim.state.state_spect.get_var
        b_fft = get_var("b_fft")
        vx_fft = get_var("vx_fft")
        vy_fft = get_var("vy_fft")
        vz_fft = get_var("vz_fft")
        nrj_A = compute_energy_from_1field_with_coef(
            b_fft, 1.0 / self.sim.params.N ** 2
        )
        nrj_K = compute_energy_from_3fields(vx_fft, vy_fft, vz_fft)
        return nrj_A + nrj_K

    def compute_energies(self):
        energies_fft = self.compute_energies_fft()
        return tuple(self.sum_wavenumbers(e_fft) for e_fft in energies_fft)

    def compute_energy(self):
        energy_fft = self.compute_energy_fft()
        return self.sum_wavenumbers(energy_fft)



"""
FrequencySpectra (:mod:`fluidsim.solvers.ns3d.output.temporal_spectra`)
==============================================================================


Provides:

.. autoclass:: TemporalSpectra3D
   :members:
   :private-members:

.. autoclass:: TemporalSpectra2D
   :members:
   :private-members:

"""

from pathlib import Path
from logging import warn

from math import pi
import numpy as np
from scipy import signal
import h5py
from rich.progress import Progress, track

from fluiddyn.util import mpi
from fluidsim.base.output.base import SpecificOutput
from fluidsim.base.output.spatiotemporal_spectra import (
    filter_tmins_paths,
    get_arange_minmax,
)


class TemporalSpectra3D(SpecificOutput):
    """
    Computes the temporal spectra.
    """

    _tag = "temporal_spectra"
    nb_dim = 3

    @classmethod
    def _complete_params_with_default(cls, params):
        tag = "temporal_spectra"

        params.output.periods_save._set_attrib(tag, 0)

        attribs = {
            "probes_deltax": 0.1,  # m
            "probes_deltay": 0.1,  # m
            "probes_region": None,  # m
            "file_max_size": 10.0,  # MB
            "SAVE_AS_FLOAT32": True,
        }

        if cls.nb_dim == 3:
            attribs["probes_deltaz"] = 0.1  # m

        params.output._set_child(
            tag,
            attribs=attribs,
        )

        params.output.temporal_spectra._set_doc(
            """
            probes_deltax, probes_deltay and probes_deltaz: float (default: 0.1)

                Probes spacing in the x, y and z directions (in params.oper.Li unit).

            probes_region: tuple (default:None)

                Boundaries of the region in the simulation domain were probes are set.

                probes_region = (xmin, xmax, ymin, ymax, zmin, zmax), in params.oper.Lx unit.

                If None, set to the whole simulation domain.

            file_max_size: float (default: 10.0)

                Maximum size of one time series file, in megabytes.

            SAVE_AS_FLOAT32: bool (default: True)

                If set to False, probes data is saved as float64.

                Warning : saving as float64 reduces digital noise at high frequencies, but double the size of the output!

            """
        )

    def __init__(self, output):
        params = output.sim.params
        try:
            params_tspec = params.output.temporal_spectra
        except AttributeError:
            warn(
                "Cannot initialize temporal spectra output because "
                "`params` does not contain parameters for this class."
            )
            return

        super().__init__(
            output,
            period_save=params.output.periods_save.temporal_spectra,
        )

        oper = self.sim.oper

        # Parameters
        self.probes_deltax = params_tspec.probes_deltax
        self.probes_deltay = params_tspec.probes_deltay
        if self.nb_dim == 3:
            self.probes_deltaz = params_tspec.probes_deltaz

        self.period_save = params.output.periods_save.temporal_spectra

        self.path_dir = Path(self.sim.output.path_run) / "probes"
        self.keys_fields = self.sim.info_solver.classes.State.keys_state_phys

        if not output._has_to_save:
            self.period_save = 0.0
        if self.period_save == 0.0:
            return

        if params_tspec.probes_region is not None:
            self.probes_region = params_tspec.probes_region
            if self.nb_dim == 3:
                xmin, xmax, ymin, ymax, zmin, zmax = self.probes_region
            else:
                xmin, xmax, ymin, ymax = self.probes_region
                zmin, zmax = 0.0, 0.0
        else:
            xmin = ymin = zmin = 0.0
            xmax = oper.Lx
            ymax = oper.Ly

            if self.nb_dim == 3:
                zmax = oper.Lz
                self.probes_region = xmin, xmax, ymin, ymax, zmin, zmax
            else:
                zmax = 0.0
                self.probes_region = xmin, xmax, ymin, ymax

        self.file_max_size = params_tspec.file_max_size
        self.SAVE_AS_FLOAT32 = params_tspec.SAVE_AS_FLOAT32

        if self.SAVE_AS_FLOAT32:
            size_1_number = 4e-6
            self.datatype = np.float32
        else:
            size_1_number = 8e-6
            self.datatype = np.float64

        if self.nb_dim == 3:
            X, Y, Z = oper.get_XYZ_loc()
        else:
            X = oper.X
            Y = oper.Y

        # round probes positions to gridpoints
        # probes spacing should at least be oper grid spacing
        self.probes_deltax = max(
            oper.deltax, oper.deltax * round(self.probes_deltax / oper.deltax)
        )
        xmin = oper.deltax * round(xmin / oper.deltax)

        self.probes_deltay = max(
            oper.deltay, oper.deltay * round(self.probes_deltay / oper.deltay)
        )
        ymin = oper.deltay * round(ymin / oper.deltay)

        if self.nb_dim == 3:
            self.probes_deltaz = max(
                oper.deltaz, oper.deltaz * round(self.probes_deltaz / oper.deltaz)
            )
            zmin = oper.deltaz * round(zmin / oper.deltaz)
        else:
            self.probes_deltaz = None
            zmin = 0.0

        # make sure probes region is not empty, and xmax is included
        xmax += 1e-15
        ymax += 1e-15
        zmax += 1e-15

        # global probes coordinates
        self.probes_x_seq = np.arange(xmin, xmax, self.probes_deltax)
        self.probes_y_seq = np.arange(ymin, ymax, self.probes_deltay)
        self.probes_z_seq = np.arange(zmin, zmax, self.probes_deltaz)

        # data directory
        if mpi.rank == 0:
            self.path_dir.mkdir(exist_ok=True)
        if mpi.nb_proc > 1:
            mpi.comm.barrier()

        # check for existing files
        paths = sorted(self.path_dir.glob("rank*.h5"))
        if paths:
            # check values in files
            with h5py.File(paths[0], "r") as file:
                if file.attrs["nb_proc"] != mpi.nb_proc:
                    raise ValueError("process number is different from files")
                if not (
                    np.allclose(file["probes_x_seq"][:], self.probes_x_seq)
                    and np.allclose(file["probes_y_seq"][:], self.probes_y_seq)
                    and np.allclose(file["probes_z_seq"][:], self.probes_z_seq)
                ):
                    raise ValueError("probes position are different from files")
            # init from files
            paths = [p for p in paths if p.name.startswith(f"rank{mpi.rank:05}")]
            if paths:
                self.path_file = paths[-1]
                with h5py.File(self.path_file, "r") as file:
                    self.index_file = file.attrs["index_file"]
                    self.probes_x_loc = file["probes_x_loc"][:]
                    self.probes_y_loc = file["probes_y_loc"][:]
                    self.probes_z_loc = file["probes_z_loc"][:]
                    self.probes_ix_loc = file["probes_ix_loc"][:]
                    self.probes_iy_loc = file["probes_iy_loc"][:]
                    self.probes_iz_loc = file["probes_iz_loc"][:]
                    self.probes_nb_loc = self.probes_x_loc.size
                    self.number_times_in_file = file["times"].size
                    self.t_last_save = file["times"][-1]
            else:
                # no probes in proc
                self.path_file = None
                self.index_file = 0
                self.number_times_in_file = 0
                self.probes_nb_loc = 0
                self.probes_x_loc = []
                self.probes_y_loc = []
                self.probes_z_loc = []
                self.probes_ix_loc = []
                self.probes_iy_loc = []
                self.probes_iz_loc = []

        else:
            # no files were found : initialize from params
            # local probes coordinates
            self.probes_x_loc = self.probes_x_seq[
                (self.probes_x_seq >= X.min()) & (self.probes_x_seq <= X.max())
            ]
            self.probes_y_loc = self.probes_y_seq[
                (self.probes_y_seq >= Y.min()) & (self.probes_y_seq <= Y.max())
            ]

            if self.nb_dim == 3:
                self.probes_z_loc = self.probes_z_seq[
                    (self.probes_z_seq >= Z.min())
                    & (self.probes_z_seq <= Z.max())
                ]
            else:
                self.probes_z_loc = self.probes_z_seq

            if self.nb_dim == 2:
                assert self.probes_z_loc.size == 1

            self.probes_nb_loc = (
                self.probes_x_loc.size
                * self.probes_y_loc.size
                * self.probes_z_loc.size
            )

            # local probes indices
            self.probes_ix_loc = np.empty(self.probes_nb_loc, dtype=int)
            self.probes_iy_loc = np.empty_like(self.probes_ix_loc)
            self.probes_iz_loc = np.zeros_like(self.probes_ix_loc)
            probe_i = 0
            for probe_x in self.probes_x_loc:
                for probe_y in self.probes_y_loc:
                    for probe_z in self.probes_z_loc:
                        probe_ix = int((probe_x - X.min()) / oper.deltax)
                        probe_iy = int((probe_y - Y.min()) / oper.deltay)
                        self.probes_ix_loc[probe_i] = probe_ix
                        self.probes_iy_loc[probe_i] = probe_iy
                        if self.nb_dim == 3:
                            probe_iz = int((probe_z - Z.min()) / oper.deltaz)
                            self.probes_iz_loc[probe_i] = probe_iz
                        probe_i += 1

            self.probes_x_loc = self._get_data_probe_from_field(X)
            self.probes_y_loc = self._get_data_probe_from_field(Y)
            if self.nb_dim == 3:
                self.probes_z_loc = self._get_data_probe_from_field(Z)

            # initialize files
            self.index_file = 0
            self.number_times_in_file = 0
            self.t_last_save = -self.period_save
            if self.probes_nb_loc > 0:
                self._init_new_file(tmin_file=self.sim.time_stepping.t)

        # size of a single write: nb_fields * probes_nb_loc + time
        probes_write_size = (
            len(self.keys_fields) * self.probes_nb_loc + 1
        ) * size_1_number
        self.max_number_times_in_file = int(
            self.file_max_size / probes_write_size
        )

    def _init_files(self, arrays_1st_time=None):
        # we don't want to do anything when this function is called.
        pass

    def _init_new_file(self, tmin_file=None):
        """Initializes a new file"""
        if tmin_file is not None:
            # max number of digits = int(log10(t_end)) + 1
            # add .3f precision = 4 additional characters
            # +2 by anticipation of potential restarts
            str_width = int(np.log10(self.sim.params.time_stepping.t_end)) + 7
            ind_str = f"tmin{tmin_file:0{str_width}.3f}"
        else:
            ind_str = f"file{self.index_file:04}"
        self.path_file = self.path_dir / f"rank{mpi.rank:05}_{ind_str}.h5"
        with h5py.File(self.path_file, "w") as file:
            file.attrs["nb_proc"] = mpi.nb_proc
            file.attrs["index_file"] = self.index_file
            file.attrs["period_save"] = self.period_save
            create_ds = file.create_dataset
            create_ds("probes_x_seq", data=self.probes_x_seq)
            create_ds("probes_y_seq", data=self.probes_y_seq)
            create_ds("probes_z_seq", data=self.probes_z_seq)

            create_ds("probes_x_loc", data=self.probes_x_loc)
            create_ds("probes_y_loc", data=self.probes_y_loc)
            create_ds("probes_z_loc", data=self.probes_z_loc)

            create_ds("probes_ix_loc", data=self.probes_ix_loc)
            create_ds("probes_iy_loc", data=self.probes_iy_loc)
            create_ds("probes_iz_loc", data=self.probes_iz_loc)

            create_ds("times", (1,), maxshape=(None,))

            for key in self.keys_fields:
                create_ds(
                    f"probes_{key}_loc",
                    (self.probes_nb_loc, 1),
                    maxshape=(self.probes_nb_loc, None),
                    dtype=self.datatype,
                )

    def _write_to_file(self, data):
        """Writes a file with the temporal data"""
        with h5py.File(self.path_file, "a") as file:
            for k, v in data.items():
                dset = file[k]
                if k.startswith("times"):
                    dset.resize((self.number_times_in_file,))
                    if self.SAVE_AS_FLOAT32:
                        v = np.array(v, dtype="float32")
                    dset[-1] = v
                else:
                    dset.resize((self.probes_nb_loc, self.number_times_in_file))
                    if self.SAVE_AS_FLOAT32:
                        v = v.astype("float32")
                    dset[:, -1] = v

    def _get_data_probe_from_field(self, field):
        return field[self.probes_iz_loc, self.probes_iy_loc, self.probes_ix_loc]

    def _add_probes_data_to_dict(self, data_dict, key):
        """Probes fields and append data to a dict object"""
        data_dict[f"probes_{key}_loc"] = self._get_data_probe_from_field(
            self.sim.state.get_var(key)
        )

    def _online_save(self):
        """Prepares data and writes to file"""
        if self.probes_nb_loc > 0:
            tsim = self.sim.time_stepping.t
            if (
                tsim + 1e-15
            ) // self.period_save > self.t_last_save // self.period_save:
                # if max write number is reached, init new file
                if self.number_times_in_file >= self.max_number_times_in_file:
                    self.index_file += 1
                    self.number_times_in_file = 0
                    self._init_new_file(tmin_file=self.sim.time_stepping.t)
                # get data from probes
                data = {"times": self.sim.time_stepping.t}
                data["times"] = self.sim.time_stepping.t
                for key in self.keys_fields:
                    self._add_probes_data_to_dict(data, key)
                # write to file
                self.number_times_in_file += 1
                self._write_to_file(data)
                self.t_last_save = tsim

    def load_time_series(
        self, keys=None, region=None, tmin=0, tmax=None, dtype=None
    ):
        """load time series from files"""
        if keys is None:
            keys = self.keys_fields
        if region is None:
            region = self._get_default_region()
        if tmax is None:
            tmax = self.sim.params.time_stepping.t_end

        if self.nb_dim == 3:
            xmin, xmax, ymin, ymax, zmin, zmax = region
        else:
            xmin, xmax, ymin, ymax = region

        # get ranks
        paths = sorted(self.path_dir.glob("rank*.h5"))
        ranks = sorted({int(p.name[4:9]) for p in paths})

        # get times from the files of first rank
        print(f"load times series...")
        paths_1st_rank = [
            p for p in paths if p.name.startswith(f"rank{ranks[0]:05}")
        ]

        if dtype is None:
            with h5py.File(paths_1st_rank[0], "r") as file:
                dtype = file[f"probes_{keys[0]}_loc"].dtype

        # get list of useful files, from tmin
        tmins_files = np.array([float(p.name[14:-3]) for p in paths_1st_rank])
        tmins_files, paths_1st_rank = filter_tmins_paths(
            tmin, tmins_files, paths_1st_rank
        )

        with Progress() as progress:
            npaths = len(paths_1st_rank)
            task_files = progress.add_task(
                "Getting times from rank 0...", total=npaths
            )

            times = []
            for ip, path in enumerate(paths_1st_rank):
                with h5py.File(path, "r") as file:
                    if tmins_files[ip] > tmax:
                        progress.update(task_files, completed=npaths)
                        break
                    times_file = file["times"][:]
                    cond_times = (times_file >= tmin) & (times_file <= tmax)
                    times.append(times_file[cond_times])
                    progress.update(task_files, advance=1)

        times = np.concatenate(times)

        tmin = times.min()
        tmax = times.max()
        print(f"tmin={tmin:8.6g}, tmax={tmax:8.6g}, nit={times.size}")

        # load series
        series = {f"probes_{k}_loc": [] for k in keys}
        with Progress() as progress:
            task_ranks = progress.add_task("Rearranging...", total=len(ranks))
            task_files = progress.add_task("Rank 00000...", total=npaths)
            # loop on all files, rank by rank
            for rank in ranks:
                paths_rank = [
                    p for p in paths if p.name.startswith(f"rank{rank:05}")
                ]

                # get list of useful files, from tmin
                tmins_files = np.array([float(p.name[14:-3]) for p in paths_rank])
                tmins_files, paths_rank = filter_tmins_paths(
                    tmin, tmins_files, paths_rank
                )

                npaths = len(paths_rank)
                progress.update(
                    task_files,
                    description=f"Rank {rank:05}...",
                    total=npaths,
                    completed=0,
                )

                # for a given rank, paths are sorted by time
                data = {f"probes_{k}_loc": [] for k in keys}
                for path_file in paths_rank:
                    # break after the last useful file
                    if tmins_files[ip] > tmax:
                        progress.update(task_files, completed=npaths)
                        break

                    with h5py.File(path_file, "r") as file:
                        probes_x = file["probes_x_loc"][:]
                        probes_y = file["probes_y_loc"][:]
                        probes_z = file["probes_z_loc"][:]

                        cond_region = (
                            (probes_x >= xmin)
                            & (probes_x <= xmax)
                            & (probes_y >= ymin)
                            & (probes_y <= ymax)
                        )

                        if self.nb_dim == 3:
                            cond_region = (
                                cond_region
                                & (probes_z >= zmin)
                                & (probes_z <= zmax)
                            )

                        cond_region = np.where(cond_region)[0]

                        for key in keys:
                            skey = f"probes_{key}_loc"
                            tmp = file[skey][cond_region, :]

                            times_file = file["times"][:]
                            its_file = get_arange_minmax(times_file, tmin, tmax)
                            data[skey].append(tmp[:, its_file])

                    # update rich task
                    progress.update(task_files, advance=1)

                for key in keys:
                    skey = f"probes_{key}_loc"
                    series[skey].append(np.concatenate(data[skey], axis=1))

                # update rich task
                progress.update(task_ranks, advance=1)

        series["times"] = times
        return series

    def _compute_spectrum(self, data):
        if not hasattr(self, "f_sample"):
            paths = sorted(self.path_dir.glob("rank*.h5"))
            with h5py.File(paths[0], "r") as file:
                self.f_sample = 1.0 / file.attrs["period_save"]
            self.domega = 2 * pi * self.f_sample / data.shape[-1]

        # TODO: I'm not sure if detrend=False is good in prod, but it's much
        # better for testing
        freq, spectrum = signal.periodogram(
            data,
            fs=self.f_sample,
            scaling="spectrum",
            detrend=False,
            return_onesided=False,
        )
        return freq, spectrum / self.domega

    def compute_spectra(self, region=None, tmin=0, tmax=None, dtype=None):
        """compute temporal spectra from files"""
        if region is None:
            region = self._get_default_region()
        if tmax is None:
            tmax = self.sim.params.time_stepping.t_end

        spectra = {"region": region, "tmin": tmin, "tmax": tmax}

        # load data
        series = self.load_time_series(
            region=region, tmin=tmin, tmax=tmax, dtype=dtype
        )

        # compute periodograms and average
        for key in series.keys():
            if key.startswith("probes_"):
                freq, spectrum = self._compute_spectrum(
                    np.concatenate(series[key])
                )
                spectrum = spectrum.mean(0)
                # get one-sided spectra
                nomega = freq.size // 2 + 1
                spectrum_onesided = np.zeros(nomega)
                spectrum_onesided[0] = spectrum[0]
                spectrum_onesided[1:] = (
                    spectrum[1:nomega] + spectrum[-1:-nomega:-1]
                )

                spectra["spectrum_" + key[7:-4]] = spectrum_onesided

        spectra["omegas"] = 2 * pi * freq[:nomega]

        # total kinetic energy
        if self.nb_dim == 3:
            spectra["spectrum_K"] = 0.5 * (
                spectra["spectrum_vx"]
                + spectra["spectrum_vy"]
                + spectra["spectrum_vz"]
            )
        else:
            spectra["spectrum_K"] = 0.5 * (
                spectra["spectrum_ux"] + spectra["spectrum_uy"]
            )

        # potential energy
        try:
            N = self.sim.params.N
            spectra["spectrum_A"] = 0.5 / N ** 2 * spectra["spectrum_b"]
        except AttributeError:
            pass

        return spectra

    def _get_default_region(self):
        p_oper = self.sim.params.oper
        return (0, p_oper.Lx, 0, p_oper.Ly, 0, p_oper.Lz)

    def plot_spectra(
        self, key=None, region=None, tmin=0, tmax=None, dtype=None, xscale="log"
    ):
        """plot temporal spectra from files"""
        if key is None:
            key = self.keys_fields[0]
        if region is None:
            region = self._get_default_region()
        if tmax is None:
            tmax = self.sim.params.time_stepping.t_end

        # load or compute spectra
        path_file = self._get_path_saved_spectra(region, tmin, tmax, dtype)
        if path_file.exists():
            spectra = self.load_spectra(
                region=region, tmin=tmin, tmax=tmax, dtype=dtype
            )
        else:
            spectra = self.save_spectra(
                region=region, tmin=tmin, tmax=tmax, dtype=dtype
            )

        # plot
        fig, ax = self.output.figure_axe()
        ax.set_xlabel(r"$\omega$")
        ax.set_ylabel("spectrum")
        ax.set_xscale(xscale)
        ax.set_yscale("log")
        ax.set_title(
            f"{key} temporal spectrum (tmin={tmin:.3f}, tmax={tmax:.3f})\n"
            + self.output.summary_simul
        )

        # specific to strat
        try:
            N = self.sim.params.N
        except AttributeError:
            omegas = spectra["omegas"]
            ax.plot(
                spectra["omegas"],
                spectra["spectrum_" + key],
                "k",
                linewidth=2,
            )
        else:
            # kinetic/potential decomposition
            EK = spectra["spectrum_K"]
            EA = spectra["spectrum_A"]
            omegas = spectra["omegas"] / N
            EKN = EK[abs(omegas - 1).argmin()]  # value @N

            ax.plot(omegas, EK, "r", linewidth=2, label=r"$E_K$")
            ax.plot(omegas, EA, "b", linewidth=2, label=r"$E_A$")
            ax.set_title(
                f"kinetic/potential energy spectrum (tmin={tmin:.3f}, tmax={tmax:.3f})\n"
                + self.output.summary_simul
            )

            # resonant modes
            if self.nb_dim == 3:
                aspect_ratio = self.sim.oper.Lx / self.sim.oper.Lz
            else:
                aspect_ratio = self.sim.oper.Lx / self.sim.oper.Ly

            def modes(nx, nz):
                return np.sqrt(nx ** 2 / (nx ** 2 + aspect_ratio ** 2 * nz ** 2))

            nxs = np.arange(1, 11)
            modes_nz1 = modes(nxs, 1)
            modes_nz2 = modes(nxs, 2)
            modes_y = np.full_like(modes_nz1, fill_value=100 * EKN)

            ax.plot(modes_nz1, modes_y, "o", label="modes $n_z=1$")
            ax.plot(modes_nz2, modes_y * 3, "o", label="modes $n_z=2$")

            # omega^-2 scaling
            omegas_scaling = np.arange(0.4, 1 + 1e-15, 0.01)
            scaling_y = EKN * omegas_scaling ** -2

            ax.plot(omegas_scaling, scaling_y, "k--")

            # eye guide @N
            ymin = EKN / 10
            _, ymax = ax.get_ylim()
            ax.vlines(1, ymin, ymax, linestyle="dotted")

            # eye guide @omega_f (specific to watu_coriolis)
            if self.sim.params.forcing.type == "watu_coriolis":
                omega_f = self.sim.params.forcing.watu_coriolis.omega_f
                ax.vlines(omega_f / N, ymin, ymax, linestyle="dotted")

            ax.set_xlabel(r"$\omega/N$")
            ax.set_ylim(ymin, ymax)
            ax.set_xlim(omegas[1], 1.5)

            ax.legend()

    def save_data_as_phys_fields(self, delta_index_times=1):
        """load temporal data and save as phys_fields array"""

        # path to saving directory
        path_dir_save = self.path_dir / "phys_fields"
        path_dir_save.mkdir(exist_ok=True)

        # get ranks
        paths = sorted(self.path_dir.glob("rank*.h5"))
        ranks = sorted({int(p.name[4:9]) for p in paths})

        # get times and probes positions from the files of first rank
        print("save probe data as 3D arrays")

        paths_1st_rank = [
            p for p in paths if p.name.startswith(f"rank{ranks[0]:05}")
        ]

        with h5py.File(paths_1st_rank[0], "r") as file:
            probes_x_seq = file["probes_x_seq"][:]
            probes_y_seq = file["probes_y_seq"][:]
            probes_z_seq = file["probes_z_seq"][:]

        times = []
        for path in paths_1st_rank:
            with h5py.File(path, "r") as file:
                times.append(file["times"][:])
        times = np.concatenate(times)[::delta_index_times]

        print(f"tmin={times.min():8.6g}, tmax={times.max():8.6g}")

        # time string width
        # digits for integer part : int(log10(t)) + 1
        # add 2 zeros, coma and 3 decimals : + 6
        width = int(np.log10(times.max())) + 7

        # probes positions
        xmin = probes_x_seq.min()
        deltax = probes_x_seq[1] - xmin
        ymin = probes_y_seq.min()
        deltay = probes_y_seq[1] - ymin
        zmin = probes_z_seq.min()
        deltaz = probes_z_seq[1] - zmin
        probes_Z, probes_Y, probes_X = np.meshgrid(
            probes_z_seq, probes_y_seq, probes_x_seq, indexing="ij"
        )

        # loop on times
        for time in track(times, description="Rearranging..."):
            # initialize arrays
            dict_arrays = {k: np.empty_like(probes_X) for k in self.keys_fields}

            # loop on ranks
            for rank in ranks:
                for path_file in paths:
                    if not path_file.name.startswith(f"rank{rank:05}"):
                        continue
                    with h5py.File(path_file, "r") as file:
                        # check if the file contains the time we're looking for
                        tmin_file, tmax_file = file["times"][[0, -1]]
                        if (time < tmin_file) or (time > tmax_file):
                            continue

                        # time index
                        it = np.where(file["times"][:] == time)[0]

                        # get global probes indices
                        coord_loc = file["probes_x_loc"][:]
                        ix = np.rint((coord_loc - xmin) / deltax).astype("int")
                        coord_loc = file["probes_y_loc"][:]
                        iy = np.rint((coord_loc - ymin) / deltay).astype("int")
                        coord_loc = file["probes_z_loc"][:]
                        iz = np.rint((coord_loc - zmin) / deltaz).astype("int")

                        # load data at time t for all keys_fields
                        for key in self.keys_fields:
                            dict_arrays[key][iz, iy, ix] = file[
                                f"probes_{key}_loc"
                            ][:, it].transpose()

                        # stop opening files when we've reached the right one
                        break

            # save fields into a new file
            path_file_save = (
                path_dir_save / f"probes_fields_t{time:0{width}.3f}.h5"
            )
            with h5py.File(path_file_save, "w") as file:
                create_ds = file.create_dataset
                # probes coordinates
                create_ds("x", data=probes_X)
                create_ds("y", data=probes_Y)
                create_ds("z", data=probes_Z)
                # physical fields
                for k, v in dict_arrays.items():
                    create_ds(k, data=v)
                # sim info
                self.sim.info._save_as_hdf5(hdf5_parent=file)

    def _get_path_saved_spectra(self, region, tmin, tmax, dtype):
        base = (
            "periodogram_" + "_".join(str(x) for x in region) + f"_{tmin}_{tmax}"
        )
        if dtype is not None:
            base += f"_{dtype}"
        return self.path_dir / (base + ".h5")

    def save_spectra(self, region=None, tmin=0, tmax=None, dtype=None):
        """compute temporal spectra from files"""
        if region is None:
            region = self._get_default_region()
        if tmax is None:
            tmax = self.sim.params.time_stepping.t_end

        spectra = self.compute_spectra(
            region=region, tmin=tmin, tmax=tmax, dtype=dtype
        )

        path_file = self._get_path_saved_spectra(region, tmin, tmax, dtype)
        with h5py.File(path_file, "w") as file:
            file.attrs["region"] = region
            file.attrs["tmin"] = tmin
            file.attrs["tmax"] = tmax
            for key, val in spectra.items():
                file.create_dataset(key, data=val)

        return spectra

    def load_spectra(self, region=None, tmin=0, tmax=None, dtype=None):
        """load temporal spectra from file"""
        if region is None:
            region = self._get_default_region()
        if tmax is None:
            tmax = self.sim.params.time_stepping.t_end

        spectra = {"region": region, "tmin": tmin, "tmax": tmax}

        path_file = self._get_path_saved_spectra(region, tmin, tmax, dtype)
        with h5py.File(path_file, "r") as file:
            for key in file.keys():
                spectra[key] = file[key][...]

        return spectra


class TemporalSpectra2D(TemporalSpectra3D):
    nb_dim = 2

    def _get_data_probe_from_field(self, field):
        return field[self.probes_iy_loc, self.probes_ix_loc]

    def save_data_as_phys_fields(self, delta_index_times=1):
        """load temporal data and save as phys_fields array"""
        raise NotImplementedError

    def _get_default_region(self):
        p_oper = self.sim.params.oper
        return (0, p_oper.Lx, 0, p_oper.Ly)



"""Utilities for benchmarking and profiling (:mod:`fluidsim.util.bench.util`)
=============================================================================

"""
import os
from time import time
from pathlib import Path

try:
    from time import perf_counter as clock
except ImportError:
    # python 2.7
    from time import clock
import json
import socket
import shutil

from fluiddyn.util import time_as_str
from fluiddyn.util import mpi
from fluiddyn.io import stdout_redirected
from ..util import available_solver_keys, get_dim_from_solver_key


old_print = print
print = mpi.printby0


class ConsoleError(ValueError):
    """Distinguish errors from console utilities."""

    pass


def modif_box_size(params, n0, n1, n2=None):
    """Modify box size, such that the aspect ratio is square / cube

    Parameters
    ----------
    params : ParamContainer
        Input parameters
    n0 : int
        Number of grid points in 0-axis of the grid
    n1 : int
        Number of grid points in 1-axis of the grid
    n2 :
        Number of grid points in 2-axis of the grid

    Returns
    -------

    """
    if n2 is None:
        nx = n1
        ny = n0
    else:
        nx = n2
        ny = n1
        nz = n0

    if nx != ny:
        if nx < ny:
            params.oper.Ly = params.oper.Ly * ny / nx
        else:
            params.oper.Lx = params.oper.Lx * nx / ny

    if n2 is None:
        mpi.printby0(
            "nh = ({}, {}); Lh = ({}, {})".format(
                n0, n1, params.oper.Ly, params.oper.Lx
            )
        )

    if n2 is not None and n2 != n0:
        params.oper.Lz = params.oper.Lx * nz / nx

        mpi.printby0(
            "n = ({}, {}, {}); L = ({}, {}, {})".format(
                n0, n1, n2, params.oper.Lz, params.oper.Ly, params.oper.Lx
            )
        )


def modif_params2d(
    params, n0=3 * 2 ** 8, n1=None, name_run="profile", type_fft=None, it_end=20
):
    """Modify parameters for 2D benchmarks.

    Parameters
    ----------
    params : ParamContainer
        Default parameters
    n0 : int
        Number of grid points in x-axis of the grid
    n1 : int
        Number of grid points in y-axis of the grid
    name_run : str
        Short name of the run
    type_fft : str
        When set, uses a diffrent FFT type than the default

    """
    params.short_name_type_run = name_run

    if n1 is None:
        n1 = n0

    params.oper.nx = n1
    params.oper.ny = n0
    modif_box_size(params, n0, n1)

    if type_fft is not None:
        params.oper.type_fft = type_fft

    if "FLUIDSIM_NO_FLUIDFFT" not in os.environ:
        # params.oper.type_fft = 'fft2d.mpi_with_fftwmpi2d'
        pass

    if "noise" in params.init_fields.available_types:
        params.init_fields.type = "noise"

    # params.forcing.enable = True
    # params.forcing.type = "tcrandom"
    # params.forcing.nkmax_forcing = 6
    # params.forcing.nkmin_forcing = 3
    # params.forcing.forcing_rate = 1.

    params.nu_8 = 1.0
    try:
        params.f = 1.0
        params.c2 = 200.0
    except AttributeError:
        pass

    try:
        params.N = 1.0
    except AttributeError:
        pass

    params.time_stepping.deltat0 = 1.0e-6
    params.time_stepping.USE_CFL = False
    params.time_stepping.it_end = it_end
    params.time_stepping.USE_T_END = False

    params.output.periods_print.print_stdout = 0
    params.output.HAS_TO_SAVE = 0


def modif_params3d(
    params, n0=256, n1=None, n2=None, name_run="profile", type_fft=None, it_end=10
):
    """Modify parameters for 3D benchmarks.

    Parameters
    ----------
    params : ParamContainer
        Default parameters
    n0 : int
        Number of grid points in x-axis of the grid
    n1 : int
        Number of grid points in y-axis of the grid
    n2 : int
        Number of grid points in z-axis of the grid
    name_run : str
        Short name of the run
    type_fft : str
        When set, uses a diffrent FFT type than the default

    """
    params.short_name_type_run = name_run

    if n1 is None:
        n1 = n0

    if n2 is None:
        n2 = n0

    params.oper.nx = n2
    params.oper.ny = n1
    params.oper.nz = n0
    modif_box_size(params, n0, n1, n2)

    if type_fft is not None:
        params.oper.type_fft = type_fft

    if "FLUIDSIM_NO_FLUIDFFT" not in os.environ:
        # params.oper.type_fft = 'fft2d.mpi_with_fftwmpi2d'
        pass

    # params.forcing.enable = False
    # params.forcing.type = 'tcrandom'
    # params.forcing.nkmax_forcing = 5
    # params.forcing.nkmin_forcing = 4
    # params.forcing.forcing_rate = 1.

    params.nu_8 = 1.0
    try:
        params.f = 1.0
        params.c2 = 200.0
    except AttributeError:
        pass

    try:
        params.N = 1.0
    except AttributeError:
        pass

    if "noise" in params.init_fields.available_types:
        params.init_fields.type = "noise"

    params.time_stepping.deltat0 = 1.0e-4
    params.time_stepping.USE_CFL = False
    params.time_stepping.it_end = it_end
    params.time_stepping.USE_T_END = False

    params.output.periods_print.print_stdout = 0
    params.output.HAS_TO_SAVE = 0


def init_parser_base(parser):
    """Initalize argument parser with common arguments for benchmarking
    analysis and profile console tools.

    Parameters
    ----------
    parser : argparse.ArgumentParser

    """
    parser.add_argument("n0", nargs="?", type=int, default=None)
    parser.add_argument("n1", nargs="?", type=int, default=None)
    parser.add_argument("n2", nargs="?", type=int, default=None)

    parser.add_argument(
        "-s",
        "--solver",
        type=str,
        default=None,
        help="Any of the following solver keys: {}".format(
            available_solver_keys()
        ),
    )
    parser.add_argument("-d", "--dim", default=None)


def parse_args_dim(args):
    """Parse dimension argument args.dim

    Parameters
    ----------
    args : argparse.ArgumentParser

    """
    dim = args.dim
    n0 = args.n0
    n1 = args.n1
    n2 = args.n2
    solver = args.solver

    if solver is not None:
        dim = get_dim_from_solver_key(solver)

    if dim is None:
        if n0 is not None and n1 is not None and n2 is None:
            dim = "2d"
        elif n0 is not None and n1 is not None and n2 is not None:
            dim = "3d"
        else:
            raise ConsoleError(
                "Cannot determine which shape you want to use for this bench "
                "('2d' or '3d')"
            )

    if dim.lower() in ["3", "3d"]:
        if n0 is None:
            n0 = 128
        if n2 is None:
            n2 = n0
        dim = "3d"
    elif dim.lower() in ["2", "2d"]:
        dim = "2d"
        if n0 is None:
            n0 = 512
    else:
        raise ConsoleError(f"dim should not be {dim}")

    if n1 is None:
        n1 = n0

    if solver is None:
        solver = "ns2d" if dim == "2d" else "ns3d"

    args.dim = dim
    args.n0 = n0
    args.n1 = n1
    args.n2 = n2
    args.solver = solver

    return args


def get_path_file(sim, path_dir_results, name="bench", ext=".json"):
    """Generate a unique filename from simulation object."""

    path_dir_results = Path(path_dir_results)

    if mpi.rank == 0:
        path_dir_results.mkdir(parents=True, exist_ok=True)

    t_as_str = time_as_str()
    key_solver = sim.info_solver.short_name.lower()
    pid = str(os.getpid())
    nb_proc = f"np={mpi.nb_proc}"
    type_fft = sim.params.oper.type_fft.split(".")[-1].replace("_", "-")
    name_file = (
        "_".join(
            [
                "result",
                name,
                key_solver,
                sim.oper.produce_str_describing_grid(),
                nb_proc,
                type_fft,
                t_as_str + pid,
            ]
        )
        + ext
    )

    path = str(path_dir_results / name_file)
    return path, t_as_str


def bench(sim, path_dir_results):
    """Benchmark a simulation run and save the results in a JSON file.

    Parameters
    ----------
    sim : Simul
        An initialized simulation object
    path_dir_results :  str
        Directory path to save results in

    """
    path, t_as_str = get_path_file(sim, path_dir_results)
    print("running a benchmark simulation... ", end="")
    with stdout_redirected():
        t0_usr = time()
        t0_sys = clock()
        sim.time_stepping.start()
        t_elapsed_sys = clock() - t0_sys
        t_elapsed_usr = time() - t0_usr

    print(
        "done.\n{} time steps computed in {:.2f} s".format(
            sim.time_stepping.it, t_elapsed_usr
        )
    )

    if sim.oper.rank != 0:
        return

    results = {
        "t_elapsed_usr": t_elapsed_usr,
        "t_elapsed_sys": t_elapsed_sys,
        "key_solver": sim.info_solver.short_name.lower(),
        "n0": sim.oper.shapeX_seq[0],
        "n1": sim.oper.shapeX_seq[1],
        "n0_loc": sim.oper.shapeX_loc[0],
        "n1_loc": sim.oper.shapeX_loc[1],
        "k0_loc": sim.oper.shapeK_loc[0],
        "k1_loc": sim.oper.shapeK_loc[1],
        "nb_proc": mpi.nb_proc,
        "pid": os.getpid(),
        "time_as_str": t_as_str,
        "hostname": socket.gethostname(),
        "nb_iter": sim.params.time_stepping.it_end,
        "type_fft": sim.oper.type_fft,
    }

    try:
        results.update(
            {
                "n2": sim.oper.shapeX_seq[2],
                "n2_loc": sim.oper.shapeX_loc[2],
                "k2_loc": sim.oper.shapeK_loc[2],
            }
        )
    except IndexError:
        pass

    with open(path, "w") as file:
        json.dump(results, file, sort_keys=True)
        file.write("\n")

    print(f"results benchmarks saved in\n{path}\n")


def tear_down(sim):
    """Delete simulation directory."""
    if mpi.rank == 0:
        print("Cleaning up simulation.")
        shutil.rmtree(sim.output.path_run, ignore_errors=True)



"""Advection-diffusion 1D solvers (:mod:`fluidsim.solvers.ad1d`)
================================================================

Provides:

.. autosummary::
   :toctree:

   solver
   state
   init_fields

"""



import h5py

import numpy as np
import matplotlib as mpl

from fluidsim.base.output.spectra3d import Spectra


def _get_averaged_spectrum(key, h5file, imin_plot, imax_plot):
    dset_spectra = h5file[key]
    spectra = dset_spectra[imin_plot : imax_plot + 1]
    spectrum = spectra.mean(0)
    spectrum[spectrum < 10e-16] = 0.0
    return spectrum


class SpectraNS3D(Spectra):
    """Save and plot spectra."""

    def compute(self):
        """compute the values at one time."""
        nrj_vx_fft, nrj_vy_fft, nrj_vz_fft = self.output.compute_energies_fft()

        s_vx_kx, s_vx_ky, s_vx_kz = self.oper.compute_1dspectra(nrj_vx_fft)
        s_vy_kx, s_vy_ky, s_vy_kz = self.oper.compute_1dspectra(nrj_vy_fft)
        s_vz_kx, s_vz_ky, s_vz_kz = self.oper.compute_1dspectra(nrj_vz_fft)

        s_kx = s_vx_kx + s_vy_kx + s_vz_kx
        s_ky = s_vx_ky + s_vy_ky + s_vz_ky
        s_kz = s_vx_kz + s_vy_kz + s_vz_kz

        dict_spectra1d = {
            "vx_kx": s_vx_kx,
            "vx_ky": s_vx_ky,
            "vx_kz": s_vx_kz,
            "vy_kx": s_vy_kx,
            "vy_ky": s_vy_ky,
            "vy_kz": s_vy_kz,
            "vz_kx": s_vz_kx,
            "vz_ky": s_vz_ky,
            "vz_kz": s_vz_kz,
            "E_kx": s_kx,
            "E_ky": s_ky,
            "E_kz": s_kz,
        }
        dict_spectra1d = {"spectra_" + k: v for k, v in dict_spectra1d.items()}

        s_vx = self.oper.compute_3dspectrum(nrj_vx_fft)
        s_vy = self.oper.compute_3dspectrum(nrj_vy_fft)
        s_vz = self.oper.compute_3dspectrum(nrj_vz_fft)

        dict_spectra3d = {
            "vx": s_vx,
            "vy": s_vy,
            "vz": s_vz,
            "E": s_vx + s_vy + s_vz,
        }
        dict_spectra3d = {"spectra_" + k: v for k, v in dict_spectra3d.items()}

        if self.has_to_save_kzkh():
            dict_kzkh = {
                "K": self.oper.compute_spectrum_kzkh(
                    nrj_vx_fft + nrj_vy_fft + nrj_vz_fft
                )
            }
        else:
            dict_kzkh = None

        return dict_spectra1d, dict_spectra3d, dict_kzkh

    def plot1d_times(
        self,
        tmin=0,
        tmax=None,
        delta_t=None,
        coef_compensate=0,
        key="E",
        key_k="kx",
        coef_plot_k3=None,
        coef_plot_k53=None,
        coef_plot_k2=None,
        xlim=None,
        ylim=None,
        only_time_average=False,
        cmap=None,
    ):

        self._plot_times(
            tmin=tmin,
            tmax=tmax,
            delta_t=delta_t,
            coef_compensate=coef_compensate,
            key=key,
            key_k=key_k,
            coef_plot_k3=coef_plot_k3,
            coef_plot_k53=coef_plot_k53,
            coef_plot_k2=coef_plot_k2,
            xlim=xlim,
            ylim=ylim,
            ndim=1,
            only_time_average=only_time_average,
            cmap=cmap,
        )

    def plot3d_times(
        self,
        tmin=0,
        tmax=None,
        delta_t=None,
        coef_compensate=0,
        key="E",
        coef_plot_k3=None,
        coef_plot_k53=None,
        coef_plot_k2=None,
        xlim=None,
        ylim=None,
        only_time_average=False,
        cmap=None,
    ):

        self._plot_times(
            tmin=tmin,
            tmax=tmax,
            delta_t=delta_t,
            coef_compensate=coef_compensate,
            key=key,
            coef_plot_k3=coef_plot_k3,
            coef_plot_k53=coef_plot_k53,
            coef_plot_k2=coef_plot_k2,
            xlim=xlim,
            ylim=ylim,
            ndim=3,
            only_time_average=only_time_average,
            cmap=cmap,
        )

    def _plot_times(
        self,
        tmin=0,
        tmax=None,
        delta_t=None,
        coef_compensate=0,
        key="E",
        key_k="kx",
        coef_plot_k3=None,
        coef_plot_k53=None,
        coef_plot_k2=None,
        xlim=None,
        ylim=None,
        only_time_average=False,
        ndim=1,
        cmap=None,
    ):

        if ndim not in [1, 3]:
            raise ValueError

        path_file = getattr(self, f"path_file{ndim}d")

        if ndim == 1:
            key_spectra = "spectra_" + key + "_" + key_k
            key_k_label = "k_" + key_k[-1]
        else:
            key_spectra = "spectra_" + key
            key_k = "k_spectra3d"
            key_k_label = "k"

        with h5py.File(path_file, "r") as h5file:
            times = h5file["times"][...]
            ks = h5file[key_k][...]

        ks_no0 = ks.copy()
        ks_no0[ks == 0] = np.nan

        if tmax is None:
            tmax = times.max()

        imin_plot = np.argmin(abs(times - tmin))
        imax_plot = np.argmin(abs(times - tmax))

        tmin_plot = times[imin_plot]
        tmax_plot = times[imax_plot]

        delta_t_save = np.diff(times).mean()

        if delta_t is None:
            nb_curves = min(20, len(times))
            delta_t = (tmax_plot - tmin_plot) / nb_curves

        delta_i_plot = int(np.round(delta_t / delta_t_save))
        if delta_i_plot == 0:
            delta_i_plot = 1
        delta_t = delta_t_save * delta_i_plot

        print(
            f"plot{ndim}d_times(tmin={tmin:8.6g}, tmax={tmax:8.6g}, delta_t={delta_t:8.6g},"
            f" coef_compensate={coef_compensate:.3f})\n"
            f"""plot {ndim}D spectra
tmin = {tmin_plot:8.6g} ; tmax = {tmax_plot:8.6g} ; delta_t = {delta_t:8.6g}
imin = {imin_plot:8d} ; imax = {imax_plot:8d} ; delta_i = {delta_i_plot}"""
        )

        fig, ax = self.output.figure_axe()
        ax.set_xlabel(f"${key_k_label}$")
        ax.set_ylabel("spectra " + key)
        ax.set_title(
            f"{ndim}D spectra (tmin={tmin:.2g}, tmax={tmax:.2g})\n"
            + self.output.summary_simul
        )
        ax.set_xscale("log")
        ax.set_yscale("log")

        if cmap is None:
            cmap = mpl.rcParams["image.cmap"]
        cmapper = getattr(mpl.pyplot.cm, cmap)
        nb_plots = imax_plot - imin_plot + 1
        colors = cmapper(np.linspace(0, 1, nb_plots))

        with h5py.File(path_file, "r") as h5file:
            dset_spectra = h5file[key_spectra]
            coef_norm = ks_no0 ** (coef_compensate)
            if not only_time_average:
                for ic, it in enumerate(
                    range(imin_plot, imax_plot + 1, delta_i_plot)
                ):
                    spectrum = dset_spectra[it]
                    spectrum[spectrum < 10e-16] = 0.0
                    ax.plot(ks, spectrum * coef_norm, color=colors[ic])

            spectra = dset_spectra[imin_plot : imax_plot + 1]
        spectrum = spectra.mean(0)
        spectrum[spectrum < 10e-16] = 0.0
        ax.plot(ks, spectrum * coef_norm, "k", linewidth=2)

        if coef_plot_k3 is not None:
            to_plot = coef_plot_k3 * ks_no0 ** (-3) * coef_norm
            ax.plot(ks[1:], to_plot[1:], "k--")

        if coef_plot_k53 is not None:
            to_plot = coef_plot_k53 * ks_no0 ** (-5.0 / 3) * coef_norm
            ax.plot(ks[1:], to_plot[1:], "k-.")

        if coef_plot_k2 is not None:
            to_plot = coef_plot_k2 * ks_no0 ** (-2) * coef_norm
            ax.plot(ks[1:], to_plot[1:], "k--")

        if xlim is not None:
            ax.set_xlim(xlim)

        if ylim is not None:
            ax.set_ylim(ylim)

    def plot_kzkh(
        self, tmin=0, tmax=None, key="K", ax=None, vmin=None, vmax=None
    ):
        data = self.load_kzkh_mean(tmin, tmax, key)
        spectrum = np.log10(data[key])
        kz = data["kz"]
        kh = data["kh_spectra"]

        if ax is None:
            fig, ax = self.output.figure_axe()
        else:
            fig = ax.figure

        ax.set_xlabel(r"$\kappa_h$")
        ax.set_ylabel("$k_z$")
        ax.set_title("log10 spectra\n" + self.output.summary_simul)

        qmesh = ax.pcolormesh(
            kh, kz, spectrum, shading="nearest", vmin=vmin, vmax=vmax
        )
        fig.colorbar(qmesh)

    def plot1d(
        self,
        tmin=0,
        tmax=None,
        coef_compensate=0,
        coef_plot_k3=None,
        coef_plot_k53=None,
        coef_plot_k2=None,
        xlim=None,
        ylim=None,
        directions=("x", "z"),
    ):

        self._plot_ndim(
            tmin,
            tmax,
            coef_compensate=coef_compensate,
            coef_plot_k3=coef_plot_k3,
            coef_plot_k53=coef_plot_k53,
            coef_plot_k2=coef_plot_k2,
            xlim=xlim,
            ylim=ylim,
            ndim=1,
            directions=directions,
        )

    def plot3d(
        self,
        tmin=0,
        tmax=None,
        coef_compensate=0,
        coef_plot_k3=None,
        coef_plot_k53=None,
        coef_plot_k2=None,
        xlim=None,
        ylim=None,
    ):

        ax = self._plot_ndim(
            tmin,
            tmax,
            coef_compensate=coef_compensate,
            coef_plot_k3=coef_plot_k3,
            coef_plot_k53=coef_plot_k53,
            coef_plot_k2=coef_plot_k2,
            xlim=xlim,
            ylim=ylim,
            ndim=3,
        )

        return ax

    def _plot_ndim(
        self,
        tmin=0,
        tmax=None,
        coef_compensate=0,
        coef_plot_k3=None,
        coef_plot_k53=None,
        coef_plot_k2=None,
        xlim=None,
        ylim=None,
        ndim=1,
        directions=("x", "z"),
    ):
        if ndim not in [1, 3]:
            raise ValueError

        path_file = getattr(self, f"path_file{ndim}d")

        if ndim == 1:
            key_k = "k" + directions[0]
            key_k_label = r",\ ".join(["k_" + letter for letter in directions])
        else:
            key_k = "k_spectra3d"
            key_k_label = "k"

        with h5py.File(path_file, "r") as h5file:
            times = h5file["times"][...]
            ks = h5file[key_k][...]

        if tmax is None:
            tmax = times.max()

        imin_plot = np.argmin(abs(times - tmin))
        imax_plot = np.argmin(abs(times - tmax))

        tmin_plot = times[imin_plot]
        tmax_plot = times[imax_plot]

        print(
            f"plot{ndim}d(tmin={tmin:8.6g}, tmax={tmax:8.6g},"
            f" coef_compensate={coef_compensate:.3f})\n"
            f"""plot {ndim}D spectra
tmin = {tmin_plot:8.6g} ; tmax = {tmax_plot:8.6g}
imin = {imin_plot:8d} ; imax = {imax_plot:8d}"""
        )

        fig, ax = self.output.figure_axe()
        ax.set_xlabel(f"${key_k_label}$")
        ax.set_ylabel("spectra")
        ax.set_title(
            f"{ndim}D spectra (tmin={tmin:.2g}, tmax={tmax:.2g})\n"
            + self.output.summary_simul
        )
        ax.set_xscale("log")
        ax.set_yscale("log")

        if ndim == 1:
            for direction in directions:
                self._plot1d_direction(
                    direction, imin_plot, imax_plot, coef_compensate, ax
                )
        else:
            ks_no0 = ks.copy()
            ks_no0[ks == 0] = np.nan
            with h5py.File(path_file, "r") as h5file:
                spectrum = h5file["spectra_E"][imin_plot : imax_plot + 1].mean(0)
                coef_norm = ks_no0 ** (coef_compensate)

            spectrum[spectrum < 10e-16] = 0.0
            ax.plot(ks, spectrum * coef_norm, "k", linewidth=2)

        ks = np.linspace(ks[1], 0.8 * ks[-1], 20)

        ks_no0 = ks.copy()
        ks_no0[ks == 0] = np.nan
        coef_norm = ks_no0 ** (coef_compensate)

        if coef_plot_k3 is not None:
            to_plot = coef_plot_k3 * ks_no0 ** (-3) * coef_norm
            ax.plot(ks[1:], to_plot[1:], "k--")

        if coef_plot_k53 is not None:
            to_plot = coef_plot_k53 * ks_no0 ** (-5.0 / 3) * coef_norm
            ax.plot(ks[1:], to_plot[1:], "k-.")

        if coef_plot_k2 is not None:
            to_plot = coef_plot_k2 * ks_no0 ** (-2) * coef_norm
            ax.plot(ks[1:], to_plot[1:], "k-.")

        if xlim is not None:
            ax.set_xlim(xlim)

        if ylim is not None:
            ax.set_ylim(ylim)

        if ndim == 1:
            fig.legend()

        return ax

    def _plot1d_direction(
        self, direction, imin_plot, imax_plot, coef_compensate, ax
    ):

        with h5py.File(self.path_file1d, "r") as h5file:
            ks = h5file["k" + direction][...]
            spectrum = _get_averaged_spectrum(
                "spectra_E_k" + direction, h5file, imin_plot, imax_plot
            )

        ks_no0 = ks.copy()
        ks_no0[ks == 0] = np.nan
        coef_norm = ks_no0 ** (coef_compensate)

        style_line = "r"
        if direction == "z":
            style_line = "g"

        ax.plot(
            ks,
            spectrum * coef_norm,
            style_line,
            linewidth=2,
            label=f"$E(k_{direction})$",
        )

    def plot3d_cumul_diss(self, tmin=0, tmax=None):
        path_file = self.path_file3d
        with h5py.File(path_file, "r") as h5file:
            times = h5file["times"][...]
        if tmax is None:
            tmax = times.max()

        # load 3D spectra
        data = self.load3d_mean(tmin=tmin, tmax=tmax)
        spectra_K = (
            data["spectra_vx"] + data["spectra_vy"] + data["spectra_vz"]
        )  # 'spectra_E' key was wrong for fluidsim<=0.3.3: use components instead
        k = data["k"]
        deltak = k[1]

        # nu_2 dissipation flux
        fd = self.sim.params.nu_2 * k ** 2
        diss = fd * spectra_K
        flux_diss_nu2 = deltak * np.cumsum(diss)

        # nu_4 dissipation flux
        fd = self.sim.params.nu_4 * k ** 4
        diss = fd * spectra_K
        flux_diss_nu4 = deltak * np.cumsum(diss)

        # normalize by total dissipation
        eps_tot = flux_diss_nu2[-1] + flux_diss_nu4[-1]
        flux_diss_nu2 /= eps_tot
        flux_diss_nu4 /= eps_tot

        # plot
        fig, ax = self.output.figure_axe()
        ax.set_xlabel(r"$k$")
        ax.set_ylabel(r"$D(k)/\epsilon$")

        ax.plot(k, flux_diss_nu2, "r-", label=r"$D_2(k)$")
        ax.plot(k, flux_diss_nu4, "m-", label=r"$D_4(k)$")

        ax.set_title(
            f"3D cumulative dissipation spectra (tmin={tmin:.2g}, tmax={tmax:.2g})\n"
            + self.output.summary_simul
        )
        ax.set_xscale("log")
        fig.legend()



import os

import numpy as np
import matplotlib.pyplot as plt

from fluiddyn.util import mpi
from fluidsim.base.output.spatial_means import SpatialMeansJSON, inner_prod
from ._old_spatial_means import load_txt as _old_load_txt


class SpatialMeansMSW1L(SpatialMeansJSON):
    """Handle the saving of spatial mean quantities.

    Viz. total energy, K.E., A.P.E. and Charney potential enstrophy. It also
    handles the computation of forcing and dissipation rates for
    sw1l.modified solver

    """

    def __init__(self, output):

        params = output.sim.params
        self.c2 = params.c2
        self.f = params.f
        if mpi.rank == 0:
            self._result = {}

        super().__init__(output)

    def _save_one_time(self):
        tsim = self.sim.time_stepping.t
        self.t_last_save = tsim

        if mpi.rank == 0:
            self._result["t"] = tsim

        (
            energyK_fft,
            energyA_fft,
            energyKr_fft,
        ) = self.output.compute_energies_fft()
        energyK = self.sum_wavenumbers(energyK_fft)
        energyA = self.sum_wavenumbers(energyA_fft)
        energyKr = self.sum_wavenumbers(energyKr_fft)
        energy = energyK + energyA

        CharneyPE_fft = self.output.compute_CharneyPE_fft()
        CharneyPE = self.sum_wavenumbers(CharneyPE_fft)

        if mpi.rank == 0:
            self._result.update(
                {
                    "E": energy,
                    "CPE": CharneyPE,
                    "EK": energyK,
                    "EA": energyA,
                    "EKr": energyKr,
                }
            )

        # Compute and save dissipation rates.
        self.treat_dissipation_rates(energyK_fft, energyA_fft, CharneyPE_fft)

        # Compute and save conversion rates.
        self.treat_conversion()

        # Compute and save skewness and kurtosis.
        eta = self.sim.state.state_phys.get_var("eta")
        meaneta2 = 2.0 / self.c2 * energyA
        if meaneta2 == 0:
            skew_eta = 0.0
            kurt_eta = 0.0
        else:
            skew_eta = np.mean(eta ** 3) / meaneta2 ** (3.0 / 2)
            kurt_eta = np.mean(eta ** 4) / meaneta2 ** (2)

        ux = self.sim.state.state_phys.get_var("ux")
        uy = self.sim.state.state_phys.get_var("uy")
        ux_fft = self.sim.oper.fft2(ux)
        uy_fft = self.sim.oper.fft2(uy)
        rot_fft = self.sim.oper.rotfft_from_vecfft(ux_fft, uy_fft)
        rot = self.sim.oper.ifft2(rot_fft)
        div_fft = self.sim.oper.divfft_from_vecfft(ux_fft, uy_fft)
        div = self.sim.oper.ifft2(div_fft)
        meanrot2 = self.sum_wavenumbers(abs(rot_fft) ** 2)
        meandiv2 = self.sum_wavenumbers(abs(div_fft) ** 2)
        if meanrot2 == 0:
            skew_rot = 0.0
            kurt_rot = 0.0
        else:
            skew_rot = np.mean(rot ** 3) / meanrot2 ** (3.0 / 2)
            kurt_rot = np.mean(rot ** 4) / meanrot2 ** (2)

        if meandiv2 == 0:
            skew_div = 0.0
            kurt_div = 0.0
        else:
            skew_div = np.mean(div ** 3) / meandiv2 ** (3.0 / 2)
            kurt_div = np.mean(div ** 4) / meandiv2 ** (2)

        if mpi.rank == 0:
            self._result.update(
                {
                    "skew_eta": skew_eta,
                    "kurt_eta": kurt_eta,
                    "skew_rot": skew_rot,
                    "kurt_rot": kurt_rot,
                    "skew_div": skew_div,
                    "kurt_div": kurt_div,
                }
            )

        if self.sim.params.forcing.enable:
            self.treat_forcing()

        if mpi.rank == 0:
            super()._save_one_time(self._result)
            self.file.flush()
            os.fsync(self.file.fileno())

        if self.has_to_plot and mpi.rank == 0:
            self.axe_a.plot(tsim, energy, "k.")
            self.axe_a.plot(tsim, energyK, "r.")
            self.axe_a.plot(tsim, energyA, "b.")

            if tsim - self.t_last_show >= self.period_show:
                self.t_last_show = tsim
                fig = self.axe_a.get_figure()
                fig.canvas.draw()

    def treat_conversion(self):
        mean_global = self.sim.oper.mean_global

        c2 = self.sim.params.c2
        eta = self.sim.state.get_var("eta")
        div = self.sim.state.get_var("div")
        h = eta + 1

        Conv = c2 / 2 * mean_global(h ** 2 * div)
        c2eta1d = c2 * mean_global(eta * div)
        c2eta2d = c2 * mean_global(eta ** 2 * div)
        c2eta3d = c2 * mean_global(eta ** 3 * div)

        if mpi.rank == 0:
            self._result.update(
                {
                    "Conv": Conv,
                    "c2eta1d": c2eta1d,
                    "c2eta2d": c2eta2d,
                    "c2eta3d": c2eta3d,
                }
            )

    def treat_dissipation_rates(self, energyK_fft, energyA_fft, CharneyPE_fft):
        """Compute and save dissipation rates."""

        f_d, f_d_hypo = self.sim.compute_freq_diss()

        dict_eps = self.compute_dissipation_rates(
            f_d, f_d_hypo, energyK_fft, energyA_fft, CharneyPE_fft
        )

        if mpi.rank == 0:
            self._result.update(dict_eps)
            self._result["epsK_tot"] = epsK_tot = (
                dict_eps["epsK"] + dict_eps["epsK_hypo"]
            )
            self._result["epsA_tot"] = epsA_tot = (
                dict_eps["epsA"] + dict_eps["epsA_hypo"]
            )
            self._result["epsCPE_tot"] = (
                dict_eps["epsCPE"] + dict_eps["epsCPE_hypo"]
            )

            if self.has_to_plot:
                tsim = self.sim.time_stepping.t
                self.axe_b.plot(tsim, epsK_tot + epsA_tot, "k.")

    def compute_dissipation_rates(
        self, f_d, f_d_hypo, energyK_fft, energyA_fft, CharneyPE_fft
    ):

        epsK = self.sum_wavenumbers(f_d * 2 * energyK_fft)
        epsK_hypo = self.sum_wavenumbers(f_d_hypo * 2 * energyK_fft)
        epsA = self.sum_wavenumbers(f_d * 2 * energyA_fft)
        epsA_hypo = self.sum_wavenumbers(f_d_hypo * 2 * energyA_fft)
        epsCPE = self.sum_wavenumbers(f_d * 2 * CharneyPE_fft)
        epsCPE_hypo = self.sum_wavenumbers(f_d_hypo * 2 * CharneyPE_fft)

        dict_eps = {
            "epsK": epsK,
            "epsK_hypo": epsK_hypo,
            "epsA": epsA,
            "epsA_hypo": epsA_hypo,
            "epsCPE": epsCPE,
            "epsCPE_hypo": epsCPE_hypo,
        }
        return dict_eps

    def get_FxFyFetafft(self):
        forcing = self.sim.forcing
        set_keys = set(self.sim.state.keys_state_spect)

        if {"ux_fft", "uy_fft", "eta_fft"} == set_keys:
            Fx_fft = forcing("ux_fft")
            Fy_fft = forcing("uy_fft")
            Feta_fft = forcing("eta_fft")
        elif {"q_fft", "ap_fft", "am_fft"} == set_keys:
            Fx_fft, Fy_fft, Feta_fft = self.sim.oper.uxuyetafft_from_qapamfft(
                forcing("q_fft"), forcing("ap_fft"), forcing("am_fft")
            )
        elif {"ap_fft", "am_fft"} == set_keys:
            Fx_fft, Fy_fft, Feta_fft = self.sim.oper.uxuyetafft_from_afft(
                forcing("ap_fft") + forcing("am_fft")
            )
        elif {"div_fft", "eta_fft", "rot_fft"} == set_keys:
            Fx_fft, Fy_fft = self.sim.oper.vecfft_from_rotdivfft(
                forcing("rot_fft"), forcing("div_fft")
            )
            Feta_fft = forcing("eta_fft")
        else:
            raise NotImplementedError(
                "Not sure how to estimate forcing rate with "
                "keys_state_spect = {}".format(set_keys)
            )

        return Fx_fft, Fy_fft, Feta_fft

    def treat_forcing(self):
        """Save forcing injection rates."""
        get_var = self.sim.state.get_var
        ux_fft = get_var("ux_fft")
        uy_fft = get_var("uy_fft")
        eta_fft = get_var("eta_fft")
        Fx_fft, Fy_fft, Feta_fft = self.get_FxFyFetafft()
        deltat = self.sim.time_stepping.deltat

        PK1_fft = inner_prod(ux_fft, Fx_fft) + inner_prod(uy_fft, Fy_fft)
        PK2_fft = deltat / 2 * (abs(Fx_fft) ** 2 + abs(Fy_fft) ** 2)

        PK1 = self.sum_wavenumbers(PK1_fft)
        PK2 = self.sum_wavenumbers(PK2_fft)

        PA1_fft = self.c2 * inner_prod(eta_fft, Feta_fft)
        PA2_fft = deltat / 2 * self.c2 * (abs(Feta_fft) ** 2)

        PA1 = self.sum_wavenumbers(PA1_fft)
        PA2 = self.sum_wavenumbers(PA2_fft)

        if mpi.rank == 0:
            PK_tot = PK1 + PK2
            PA_tot = PA1 + PA2
            self._result.update(
                {
                    "PK1": PK1,
                    "PK2": PK2,
                    "PK_tot": PK_tot,
                    "PA1": PA1,
                    "PA2": PA2,
                    "PA_tot": PA_tot,
                }
            )

        if self.has_to_plot and mpi.rank == 0:
            tsim = self.sim.time_stepping.t
            self.axe_b.plot(tsim, PK_tot + PA_tot, "c.")

    def load(self):
        if self._file_exists():
            return super().load()
        else:
            dict_results = {"name_solver": self.output.name_solver}
            path_file = self.path_file.replace(".json", ".txt")
            if os.path.exists(path_file):
                return _old_load_txt(path_file, dict_results)
            else:
                raise FileNotFoundError(
                    f"Neither {self.path_file} nor {os.path.basename(path_file)}"
                    "exists"
                )

    def plot(self):
        dict_results = self.load()

        t = dict_results["t"]

        E = dict_results["E"]
        CPE = dict_results["CPE"]

        EK = dict_results["EK"]
        EA = dict_results["EA"]
        EKr = dict_results["EKr"]

        epsK = dict_results["epsK"]
        epsK_hypo = dict_results["epsK_hypo"]
        epsK_tot = dict_results["epsK_tot"]

        epsA = dict_results["epsA"]
        epsA_hypo = dict_results["epsA_hypo"]
        epsA_tot = dict_results["epsA_tot"]

        epsE = epsK + epsA
        epsE_hypo = epsK_hypo + epsA_hypo
        epsE_tot = epsK_tot + epsA_tot

        epsCPE = dict_results["epsCPE"]
        epsCPE_hypo = dict_results["epsCPE_hypo"]
        epsCPE_tot = dict_results["epsCPE_tot"]

        if "PK_tot" in dict_results:
            PK_tot = dict_results["PK_tot"]
            PA_tot = dict_results["PA_tot"]
            P_tot = PK_tot + PA_tot

        width_axe = 0.85
        height_axe = 0.37
        x_left_axe = 0.12
        z_bottom_axe = 0.56

        size_axe = [x_left_axe, z_bottom_axe, width_axe, height_axe]
        fig, ax1 = self.output.figure_axe(size_axe=size_axe)
        ax1.set_xlabel("t")
        ax1.set_ylabel("$2E(t)/c^2$")
        ax1.set_title("mean energy\n" + self.output.summary_simul)
        norm = self.c2 / 2
        ax1.plot(t, E / norm, "k", linewidth=2, label="$E$")
        ax1.plot(t, EK / norm, "r", linewidth=1, label="$E_K$")
        ax1.plot(t, EA / norm, "b", linewidth=1, label="$E_A$")
        ax1.plot(t, EKr / norm, "r--", linewidth=1, label="$E_K^r$")
        ax1.plot(t, (EK - EKr) / norm, "r:", linewidth=1, label="$E_K^d$")
        ax1.legend()

        z_bottom_axe = 0.07
        size_axe[1] = z_bottom_axe
        ax2 = fig.add_axes(size_axe)
        ax2.set_xlabel("t")
        ax2.set_ylabel("Charney PE(t)")
        ax2.set_title("mean Charney PE(t)")
        ax2.plot(t, CPE, "k", linewidth=2)

        z_bottom_axe = 0.56
        size_axe[1] = z_bottom_axe
        fig, ax1 = self.output.figure_axe(size_axe=size_axe)
        ax1.set_xlabel("t")
        ax1.set_ylabel(r"$P_E(t)$, $\epsilon(t)$")
        ax1.set_title("forcing and dissipation\n" + self.output.summary_simul)

        if "PK_tot" in dict_results:
            ax1.plot(t, P_tot, "c", linewidth=2, label="$P_{tot}$")

        ax1.plot(t, epsE, "k--", linewidth=2, label=r"$\epsilon$")
        ax1.plot(t, epsE_hypo, "g", linewidth=2, label=r"$\epsilon_{hypo}$")
        ax1.plot(t, epsE_tot, "k", linewidth=2, label=r"$\epsilon_{tot}$")

        ax1.legend(loc=2)

        z_bottom_axe = 0.07
        size_axe[1] = z_bottom_axe
        ax2 = fig.add_axes(size_axe)
        ax2.set_xlabel("t")
        ax2.set_ylabel(r"$\epsilon$ Charney PE(t)")
        ax2.set_title("dissipation Charney PE")
        ax2.plot(t, epsCPE, "k--", linewidth=2)
        ax2.plot(t, epsCPE_hypo, "g", linewidth=2)
        ax2.plot(t, epsCPE_tot, "r", linewidth=2)

    def plot_rates(self, keys="E"):
        """Plots the time history of the time derivative of a spatial mean,
        and also calculates the average of the same.

        Parameters
        ----------
        key : string or a list of strings

            Refers to the the spatial mean which you want to take time
            derivative of.  Legal value include:

            For ns2d ['E', 'Z']
            For sw1l ['E', 'EK', 'EA', 'EKr', 'CPE']

        Examples
        --------
        >>> plot_rates()
        >>> plot_rates('Z')
        >>> plot_rates(['E', 'Z'])
        >>> plot_rates(['E', 'EK', 'EA', 'EKr', 'CPE'])

        """

        dict_results = self.load()
        t = dict_results["t"]
        dt = np.gradient(t, 1.0)

        fig, axarr = plt.subplots(len(keys), sharex=True)
        i = 0
        for k in keys:
            E = dict_results[k]
            dE_dt = abs(np.gradient(E, 1.0) / dt)
            dE_dt_avg = "{:11.6e}".format(dE_dt.mean())
            try:
                axarr[i].semilogy(t, dE_dt, label=dE_dt_avg)
                axarr[i].set_ylabel(r"$\partial_t$" + keys[i])
                axarr[i].legend()
            # axarr[i].text(0.8, 0.9, 'mean = ' + dE_dt_avg, horizontalalignment='center', verticalalignment='center',)
            except TypeError:
                axarr.semilogy(t, dE_dt, label=dE_dt_avg)
                axarr.set_ylabel(keys)
                axarr.legend()
            i += 1

        try:
            axarr[i - 1].set_xlabel("t")
        except TypeError:
            axarr.set_xlabel("t")

        plt.draw()


class SpatialMeansSW1L(SpatialMeansMSW1L):
    """Handle the saving of spatial mean quantities.


    Viz. total energy, K.E., A.P.E. and Charney potential enstrophy. It also
    handles the computation of forcing and dissipation rates for sw1l solver.

    """

    def compute_dissipation_rates(
        self, f_d, f_d_hypo, energyK_fft, energyA_fft, CharneyPE_fft
    ):
        """Compute and save dissipation rates."""

        dict_eps = super().compute_dissipation_rates(
            f_d, f_d_hypo, energyK_fft, energyA_fft, CharneyPE_fft
        )

        (epsKsuppl, epsKsuppl_hypo) = self.compute_epsK(
            f_d, f_d_hypo, energyK_fft, dict_eps
        )

        dict_eps.update({"epsKsup": epsKsuppl, "epsKsup_hypo": epsKsuppl_hypo})
        return dict_eps

    def compute_epsK(self, f_d, f_d_hypo, energyK_fft, dict_eps):

        ux = self.sim.state.state_phys.get_var("ux")
        uy = self.sim.state.state_phys.get_var("uy")

        EKquad = 0.5 * (ux ** 2 + uy ** 2)
        EKquad_fft = self.sim.oper.fft2(EKquad)

        eta_fft = self.sim.state.get_var("eta_fft")

        epsKsuppl = self.sum_wavenumbers(f_d * inner_prod(EKquad_fft, eta_fft))

        epsKsuppl_hypo = self.sum_wavenumbers(
            f_d_hypo * inner_prod(EKquad_fft, eta_fft)
        )

        dict_eps["epsK"] += epsKsuppl
        dict_eps["epsK_hypo"] += epsKsuppl_hypo

        return epsKsuppl, epsKsuppl_hypo

    def treat_forcing(self):
        """
        Save forcing injection rates.
        """
        get_var = self.sim.state.get_var
        ux_fft = get_var("ux_fft")
        uy_fft = get_var("uy_fft")
        eta_fft = get_var("eta_fft")

        Fx_fft, Fy_fft, Feta_fft = self.get_FxFyFetafft()
        deltat = self.sim.time_stepping.deltat

        PA1_fft = self.c2 * inner_prod(eta_fft, Feta_fft)
        PA2_fft = deltat / 2 * self.c2 * (abs(Feta_fft) ** 2)

        PA1 = self.sum_wavenumbers(PA1_fft)
        PA2 = self.sum_wavenumbers(PA2_fft)

        Fx = self.sim.oper.ifft2(Fx_fft)
        Fy = self.sim.oper.ifft2(Fy_fft)
        Feta = self.sim.oper.ifft2(Feta_fft)

        eta = self.sim.state.state_phys.get_var("eta")
        h = eta + 1.0

        ux = self.sim.state.state_phys.get_var("ux")
        uy = self.sim.state.state_phys.get_var("uy")

        FetaFx_fft = self.sim.oper.fft2(Feta * Fx)
        FetaFy_fft = self.sim.oper.fft2(Feta * Fy)

        Jx_fft = self.sim.oper.fft2(h * ux)
        Jy_fft = self.sim.oper.fft2(h * uy)

        FJx_fft = self.sim.oper.fft2(h * Fx + Feta * ux)
        FJy_fft = self.sim.oper.fft2(h * Fy + Feta * uy)

        PK1_fft = 0.5 * (
            inner_prod(Jx_fft, Fx_fft)
            + inner_prod(Jy_fft, Fy_fft)
            + inner_prod(ux_fft, FJx_fft)
            + inner_prod(uy_fft, FJy_fft)
        )
        PK2_fft = (
            deltat
            / 2
            * (
                0.5 * (inner_prod(Fx_fft, FJx_fft) + inner_prod(Fy_fft, FJy_fft))
                + inner_prod(ux_fft, FetaFx_fft)
                + inner_prod(uy_fft, FetaFy_fft)
            )
        )

        PK1 = self.sum_wavenumbers(PK1_fft)
        PK2 = self.sum_wavenumbers(PK2_fft)

        if mpi.rank == 0:

            PK_tot = PK1 + PK2
            PA_tot = PA1 + PA2
            self._result.update(
                {
                    "PK1": PK1,
                    "PK2": PK2,
                    "PK_tot": PK_tot,
                    "PA1": PA1,
                    "PA2": PA2,
                    "PA_tot": PA_tot,
                }
            )

        if self.has_to_plot and mpi.rank == 0:
            tsim = self.sim.time_stepping.t
            self.axe_b.plot(tsim, PK_tot + PA_tot, "c.")



"""Utilities for the numerical simulations (:mod:`fluidsim.util.util`)
======================================================================

Public API
----------

.. autofunction:: load_sim_for_plot

.. autofunction:: load_state_phys_file

.. autofunction:: load_for_restart

.. autofunction:: modif_resolution_from_dir

.. autofunction:: modif_resolution_from_dir_memory_efficient

Internal API
------------

.. autofunction:: open_patient

"""

import os
import time
from copy import deepcopy as _deepcopy
from datetime import timedelta
from functools import partial
from importlib import import_module
from pathlib import Path
from time import perf_counter
from typing import Union

import fluiddyn as fld
import h5netcdf
import h5py
import numpy as _np
from fluiddyn.io.redirect_stdout import stdout_redirected
from fluiddyn.util import mpi
from fluiddyn.util.util import get_memory_usage

from fluidsim_core import loader

from fluidsim import path_dir_results

from fluidsim.base.params import (
    Parameters,
    fix_old_params,
    load_info_solver,
    load_params_simul,
    merge_params,
)
from fluidsim.base.solvers.info_base import create_info_simul
from fluidsim.extend_simul import _extend_simul_class_from_path

from .output import save_file

available_solvers = partial(
    loader.available_solvers, entrypoint_grp="fluidsim.solvers"
)

import_module_solver_from_key = partial(
    loader.import_module_solver, entrypoint_grp="fluidsim.solvers"
)

import_simul_class_from_key = partial(
    loader.import_cls_simul, entrypoint_grp="fluidsim.solvers"
)


def print_memory_usage_seq(message, flush=None):
    mem = get_memory_usage()
    print(message, f"{mem/1024: 7.3f} Go", flush=flush)


def available_solver_keys():
    """List all available solvers.

    Returns
    -------
    list

    """
    return sorted(available_solvers())


def get_dim_from_solver_key(key):
    """Try to guess the dimension from the solver key (via the operator name)."""
    cls = import_simul_class_from_key(key)
    info = cls.InfoSolver(only_root=True)
    class_name = info.classes.Operators.class_name

    # special cases:
    if class_name == "OperatorsPseudoSpectralSW1L":
        return "2"

    for dim in range(4):
        if str(dim) in class_name:
            return str(dim)

    raise NotImplementedError(
        "Cannot deduce dimension of the solver from the name " + class_name
    )


def pathdir_from_namedir(name_dir: Union[str, Path, None] = None):
    """Return the path of a result directory.

    name_dir: str, optional

      Can be an absolute path, a relative path, or even simply just
      the name of the directory under $FLUIDSIM_PATH.

    """
    if name_dir is None:
        return Path.cwd()

    if not isinstance(name_dir, Path):
        name_dir = Path(name_dir)

    name_dir = name_dir.expanduser()

    if name_dir.is_dir():
        return name_dir.absolute()

    if not name_dir.is_absolute():
        name_dir = path_dir_results / name_dir

    if not name_dir.is_dir():
        raise ValueError(str(name_dir) + " is not a directory")

    return name_dir


class ModulesSolvers(dict):
    """Dictionary to gather imported solvers."""

    def __init__(self, names_solvers):
        for key in names_solvers:
            self[key] = import_module_solver_from_key(key)


def name_file_from_time_approx(path_dir, t_approx=None):
    """Return the file name whose time is the closest to the given time.

    Parameters
    ----------

    path_dir: Path or str

      Path of the directory of the simulation.

    t_approx : number or "last" (optional)

      Approximate time of the file to be loaded.

    .. todo::

        Can be elegantly implemented using regex as done in
        ``fluidsim.base.output.phys_fields.time_from_path``

    """
    if not isinstance(path_dir, Path):
        path_dir = Path(path_dir)

    path_files = sorted(path_dir.glob("state_phys_t*"))

    nb_files = len(path_files)
    if nb_files == 0 and mpi.rank == 0:
        raise ValueError("No state file in the dir\n" + str(path_dir))

    if t_approx is None:
        # should be the last one but not 100% sure
        return path_files[-1].name

    name_files = [path.name for path in path_files]
    if "state_phys_t=" in name_files[0]:
        ind_start_time = len("state_phys_t=")
    else:
        ind_start_time = len("state_phys_t")

    times = _np.empty([nb_files])
    for ii, name in enumerate(name_files):
        tmp = ".".join(name[ind_start_time:].split(".")[:2])
        if "_" in tmp:
            tmp = tmp[: tmp.index("_")]
        times[ii] = float(tmp)
    if t_approx == "last":
        t_approx = times.max()
    i_file = abs(times - t_approx).argmin()
    name_file = path_files[i_file].name
    return name_file


def load_sim_for_plot(
    name_dir=None, merge_missing_params=False, hide_stdout=False
):
    """Create a object Simul from a dir result.

    Creating simulation objects with this function should be fast because the
    state is not initialized with the output file and only a coarse operator is
    created.

    Parameters
    ----------

    name_dir : str (optional)

      Name of the directory of the simulation. If nothing is given, we load the
      data in the current directory.
      Can be an absolute path, a relative path, or even simply just
      the name of the directory under $FLUIDSIM_PATH.

    merge_missing_params : bool (optional, default == False)

      Can be used to load old simulations carried out with an old fluidsim
      version.

    hide_stdout : bool (optional, default == False)

      If True, without stdout.

    """
    path_dir = pathdir_from_namedir(name_dir)
    solver = _import_solver_from_path(path_dir)
    params = load_params_simul(path_dir)

    if merge_missing_params:
        merge_params(params, solver.Simul.create_default_params())

    params.path_run = path_dir
    params.init_fields.type = "constant"
    params.init_fields.modif_after_init = False
    params.ONLY_COARSE_OPER = True
    params.NEW_DIR_RESULTS = False
    params.output.HAS_TO_SAVE = False
    params.output.ONLINE_PLOT_OK = False

    try:
        params.preprocess.enable = False
    except AttributeError:
        pass

    try:
        params.oper.type_fft = "default"
        params.oper.type_fft2d = "sequential"
    except AttributeError:
        pass

    fix_old_params(params)

    try:
        path_file = Path(path_dir) / name_file_from_time_approx(path_dir)
    except ValueError:
        # no state_phys file to extend the simul class. Let's try with info_solver.xml
        path_file = Path(path_dir) / "info_solver.xml"

    if path_file.exists():
        Simul = _extend_simul_class_from_path(solver.Simul, path_file)
    else:
        Simul = solver.Simul

    with stdout_redirected(hide_stdout):
        sim = Simul(params)
    return sim


def _import_solver_from_path(path_dir):
    info_solver = load_info_solver(path_dir)
    solver = import_module(info_solver.module_name)
    return solver


def load_state_phys_file(
    name_dir=None,
    t_approx="last",
    modif_save_params=True,
    merge_missing_params=False,
    init_with_initialized_state=True,
    hide_stdout=False,
):
    """Create a simulation from a file.

    For large resolution, creating a simulation object with this function can
    be slow because the state is initialized with the output file.

    Parameters
    ----------

    name_dir : str (optional)

      Name of the directory of the simulation. If nothing is given, we load the
      data in the current directory.
      Can be an absolute path, a relative path, or even simply just
      the name of the directory under $FLUIDSIM_PATH.

    t_approx : number (optional)

      Approximate time of the file to be loaded.

    modif_save_params :  bool (optional, default == True)

      If True, the parameters of the simulation are modified before loading::

        params.output.HAS_TO_SAVE = False
        params.output.ONLINE_PLOT_OK = False

    merge_missing_params : bool (optional, default == False)

      Can be used to load old simulations carried out with an old fluidsim
      version.

    init_with_initialized_state : bool (optional, default == True)

      If True, call sim.output.init_with_initialized_state.

    hide_stdout : bool (optional, default == False)

      If True, without stdout.

    """

    params, Simul = load_for_restart(name_dir, t_approx, merge_missing_params)

    if modif_save_params:
        params.output.HAS_TO_SAVE = False
        params.output.ONLINE_PLOT_OK = False

    params.ONLY_COARSE_OPER = False
    try:
        params.oper.type_fft = "default"
        params.oper.type_fft2d = "sequential"
    except AttributeError:
        pass

    with stdout_redirected(hide_stdout):
        sim = Simul(params)

    if init_with_initialized_state:
        sim.output.init_with_initialized_state()

    return sim


def load_for_restart(name_dir=None, t_approx="last", merge_missing_params=False):
    """Load params and Simul for a restart.

    >>> params, Simul = load_for_restart(name_dir)

    Parameters
    ----------

    name_dir : str (optional)

      Name of the directory of the simulation. If nothing is given, we load the
      data in the current directory.
      Can be an absolute path, a relative path, or even simply just
      the name of the directory under $FLUIDSIM_PATH.

    t_approx : number (optional)

      Approximate time of the file to be loaded.

    merge_missing_params : bool (optional, default == False)

      Can be used to load old simulations carried out with an old fluidsim
      version.

    """

    path_dir = pathdir_from_namedir(name_dir)
    solver = _import_solver_from_path(path_dir)
    Simul = solver.Simul

    # choose the file with the time closer to t_approx
    name_file = name_file_from_time_approx(path_dir, t_approx)
    path_file = os.path.join(path_dir, name_file)

    Simul = _extend_simul_class_from_path(Simul, path_file)

    if merge_missing_params:
        # this has to be done by all processes otherwise there is a problem
        # with Transonic (see https://foss.heptapod.net/fluiddyn/fluidsim/issues/26)
        default_params = Simul.create_default_params()

    if mpi.rank > 0:
        params = None
    else:
        with h5py.File(path_file, "r") as file:
            params = Parameters(hdf5_object=file["info_simul"]["params"])

        if merge_missing_params:
            merge_params(params, default_params)

        params.path_run = path_dir
        params.NEW_DIR_RESULTS = False
        params.init_fields.type = "from_file"
        params.init_fields.from_file.path = path_file
        params.init_fields.modif_after_init = False
        try:
            params.preprocess.enable = False
        except AttributeError:
            pass

        fix_old_params(params)

    if mpi.nb_proc > 1:
        params = mpi.comm.bcast(params, root=0)

    return params, Simul


def modif_resolution_all_dir(t_approx=None, coef_modif_resol=2, dir_base=None):
    """Save files with a modified resolution."""
    raise DeprecationWarning("Sorry, use modif_resolution_from_dir instead")


def modif_resolution_from_dir(
    name_dir=None, t_approx="last", coef_modif_resol=2, PLOT=True
):
    """Save a file with a modified resolution."""

    sim = load_state_phys_file(name_dir, t_approx)

    params2 = _deepcopy(sim.params)
    params2.output.HAS_TO_SAVE = True
    params2.oper.nx = int(sim.params.oper.nx * coef_modif_resol)
    params2.oper.ny = int(sim.params.oper.ny * coef_modif_resol)

    try:
        params2.oper.nz = int(sim.params.oper.nz * coef_modif_resol)
    except AttributeError:
        dimension = 2
    else:
        dimension = 3

    params2.init_fields.type = "from_simul"

    sim2 = sim.__class__(params2)
    sim2.init_fields.get_state_from_simul(sim)

    oper_new = sim2.params.oper
    if dimension == 3:
        dir_new_file = f"/State_phys_{oper_new.nx}x{oper_new.ny}x{oper_new.nz}"
    else:
        dir_new_file = f"/State_phys_{oper_new.nx}x{oper_new.ny}"

    sim2.output.path_run = str(sim.output.path_run) + dir_new_file
    print("Save file in directory\n" + sim2.output.path_run)
    sim2.output.phys_fields.save(particular_attr="modif_resolution")

    print("The new file is saved.")

    if PLOT:
        sim.output.phys_fields.plot(numfig=0)
        sim2.output.phys_fields.plot(numfig=1)
        fld.show()


class StatePhysLike:
    def __init__(self, path_file, oper, oper2):
        self.path_file = path_file
        self.oper = oper
        self.oper2 = oper2
        self.info = "state_phys"

        self.field = oper.create_arrayX()
        self.field_spect = oper.create_arrayK()
        print_memory_usage_seq("Memory usage after init fields:           ")

        self.field2 = oper2.create_arrayX()
        print(
            "size field2:                               "
            f"{self.field2.nbytes / 1024**3:7.3f} Go"
        )
        print_memory_usage_seq("Memory usage after init field2:           ")
        self.field2_spect = oper2.create_arrayK(0)
        print(
            "size field2_spect:                         "
            f"{self.field2_spect.nbytes / 1024**3:7.3f} Go"
        )
        print_memory_usage_seq("Memory usage after init field2_spect:     ")

        if path_file.suffix == ".nc":
            self.h5pack = h5netcdf
        else:
            self.h5pack = h5py

        with self.h5pack.File(self.path_file, "r") as h5file:
            group_state_phys = h5file["/state_phys"]
            self.keys = list(group_state_phys.keys())
            self.time = float(group_state_phys.attrs["time"])
            self.it = int(group_state_phys.attrs["it"])
            self.name_run = h5file.attrs["name_run"]

    def get_var(self, key):
        print(f'get_var("{key}")')

        def start_counter(message):
            print(f"- {message + '...':30s}", end="", flush=True)
            return perf_counter()

        def end_counter(t_start):
            print(f"done in {timedelta(seconds=perf_counter() - t_start)}")

        t_start = start_counter("reading field from disk")
        with self.h5pack.File(self.path_file, "r") as h5file:
            group_state_phys = h5file["/state_phys"]
            self.field[:] = group_state_phys[key][...]
        end_counter(t_start)

        t_start = start_counter("forward fft smaller field")
        self.oper.fft_as_arg(self.field, self.field_spect)
        end_counter(t_start)

        dimension = len(self.field_spect.shape)
        if dimension not in [2, 3]:
            raise NotImplementedError

        t_start = start_counter("filling field2_fft")
        self.oper2.fill_field_fft(self.field_spect, self.field2_spect, self.oper)
        end_counter(t_start)

        t_start = start_counter("backward fft field2")
        self.oper2.ifft_as_arg(self.field2_spect, self.field2)
        end_counter(t_start)

        return self.field2


def modif_resolution_from_dir_memory_efficient(
    name_dir=None, t_approx="last", coef_modif_resol=2
):
    """Save a file with a modified resolution.

    Faster and more memory efficient than ``modif_resolution_from_dir`` (but
    not plot).

    """
    t_start = perf_counter()

    if mpi.nb_proc > 1:
        raise NotImplementedError

    path_dir = pathdir_from_namedir(name_dir)
    solver = _import_solver_from_path(path_dir)

    name_file = name_file_from_time_approx(path_dir, t_approx)
    path_file = path_dir / name_file
    print(f"Changing resolution of the state contained in\n{path_file}")

    Simul = _extend_simul_class_from_path(solver.Simul, path_file)
    try:
        info_solver = Simul.info_solver
    except AttributeError:
        info_solver = Simul.InfoSolver()

    with h5py.File(path_file, "r") as h5file:
        params = Parameters(hdf5_object=h5file["/info_simul/params"])

    try:
        params.oper.type_fft = "default"
        params.oper.type_fft2d = "sequential"
    except AttributeError:
        pass

    params2 = _deepcopy(params)
    params2.output.HAS_TO_SAVE = True
    nx2 = params2.oper.nx = int(params.oper.nx * coef_modif_resol)
    ny2 = params2.oper.ny = int(params.oper.ny * coef_modif_resol)

    try:
        nz2 = params2.oper.nz = int(params.oper.nz * coef_modif_resol)
    except AttributeError:
        dimension = 2
        shape = (params.oper.ny, params.oper.nx)
        shape2 = (ny2, nx2)
    else:
        dimension = 3
        shape = (params.oper.nz, params.oper.ny, params.oper.nx)
        shape2 = (nz2, ny2, nx2)

    from .mini_oper_modif_resol import MiniOperModifResol

    oper = MiniOperModifResol(shape)
    print_memory_usage_seq(
        'Memory usage after init operator "input": ', flush=True
    )
    oper2 = MiniOperModifResol(shape2)

    print_memory_usage_seq('Memory usage after init operator "output":')
    info2 = create_info_simul(info_solver, params2)

    state_phys = StatePhysLike(path_file, oper, oper2)

    if dimension == 3:
        dir_new_new = f"State_phys_{nx2}x{ny2}x{nz2}"
    else:
        dir_new_new = f"State_phys_{nx2}x{ny2}"

    path_file_out = path_file.parent / dir_new_new / path_file.name
    path_file_out.parent.mkdir(exist_ok=True)
    print(f"Saving file {path_file_out.name}...", flush=True)
    save_file(
        path_file_out,
        state_phys,
        info2,
        state_phys.name_run,
        oper2,
        state_phys.time,
        state_phys.it,
        particular_attr="modif_resolution",
    )
    print(
        f"File {path_file_out.name} saved in:\n{path_file_out.parent}\n"
        f"total duration: {timedelta(seconds=perf_counter() - t_start)}"
    )


def times_start_end_from_path(path):
    """Return the start and end times from a result directory path."""

    path_file = path + "/stdout.txt"
    if not os.path.exists(path_file):
        print("Given path does not exist:\n " + path)
        return 666, 666

    with open(path_file, "r") as file_stdout:

        line = ""
        while not line.startswith("it ="):
            line = file_stdout.readline()

        words = line.split()
        t_s = float(words[6])

        # in order to get the information at the end of the file,
        # we do not want to read the full file...
        file_stdout.seek(0, os.SEEK_END)  # go to the end
        nb_caract = file_stdout.tell()
        nb_caract_to_read = min(nb_caract, 1000)
        file_stdout.seek(file_stdout.tell() - nb_caract_to_read, os.SEEK_SET)
        while line != "":
            if line.startswith("it ="):
                line_it = line
            last_line = line
            line = file_stdout.readline()

        if last_line.startswith("save state_phys"):
            name_file = last_line.split()[-1]
            name_file, ext = os.path.splitext(name_file)
            word = name_file.split("_it=")[0].split("state_phys_t")[-1]
            t_e = float(word.replace(ext, ""))
        else:
            words = line_it.split()
            t_e = float(words[6])

    # print('t_s = {0:.3f}, t_e = {1:.3f}'.format(t_s, t_e))

    return t_s, t_e


def open_patient(
    path,
    *args,
    time_wait_total=200,
    time_wait_once=2,
    class_file=h5py.File,
    **kwargs,
):
    """Open a hdf5 type file in a "patient" way.

    If the file is already opened by another process (``errno==11`` for hdf5),
    the error is caught and we retry later.

    Parameters
    ----------

    time_wait_total : number (optional)

      Time to wait before raising the error.

    time_wait_once : number (optional)

      Time between attempts.

    class_file : type (optional)

      Class of the file (default ``h5py.File``, but could be ``h5netcdf.File``).

    """
    time_first_try = time.time()
    while True:
        try:
            file = class_file(path, *args, **kwargs)
            break
        except OSError as error:
            errno = int(repr(error).split("errno = ")[1].split(",")[0])
            if errno != 11 or time.time() - time_first_try > time_wait_total:
                raise
            time.sleep(time_wait_once)
    return file



"""State for the NS2D solver (:mod:`fluidsim.solvers.ns2d.strat.state`)
=======================================================================

.. autoclass:: StateNS2DStrat
   :members:
   :private-members:

"""

import numpy as np

# from fluidsim.base.state import StatePseudoSpectral

from fluiddyn.util import mpi

from fluidsim.solvers.ns2d.state import StateNS2D


class StateNS2DStrat(StateNS2D):
    """State for the solver ns2d.strat.

    Contains the variables corresponding to the state and handles the
    access to other fields.

    """

    @staticmethod
    def _complete_info_solver(info_solver):
        """Update `info_solver` container with the stratification terms."""
        # Updating the state to a stratified state
        info_solver.classes.State._set_attribs(
            {
                "keys_state_spect": ["rot_fft", "b_fft"],
                "keys_state_phys": ["ux", "uy", "rot", "b"],
                "keys_computable": [
                    "ux_fft",
                    "uy_fft",
                    "div_fft",
                    "div",
                    "ap_fft",
                    "am_fft",
                    "ap",
                    "am",
                ],
                "keys_phys_needed": ["rot", "b"],
                "keys_linear_eigenmodes": ["rot_fft", "b_fft"],
            }
        )

    def __init__(self, sim, oper=None):

        super().__init__(sim, oper)

        self.field_tmp4 = np.empty_like(self.state_phys[0])
        self.field_tmp5 = np.empty_like(self.state_phys[0])

    def compute(self, key, SAVE_IN_DICT=True, RAISE_ERROR=True):
        """Compute and return a variable"""
        it = self.sim.time_stepping.it
        if key in self.vars_computed and it == self.it_computed[key]:
            return self.vars_computed[key]

        if key == "ux_fft":
            rot_fft = self.state_spect.get_var("rot_fft")
            ux_fft, uy_fft = self.oper.vecfft_from_rotfft(rot_fft)
            result = ux_fft
        # result = self.oper.fft2(self.state_phys.get_var("ux"))
        elif key == "uy_fft":
            rot_fft = self.state_spect.get_var("rot_fft")
            ux_fft, uy_fft = self.oper.vecfft_from_rotfft(rot_fft)
            result = uy_fft
        # result = self.oper.fft2(self.state_phys.get_var("uy"))
        elif key == "rot_fft":
            ux_fft = self.compute("ux_fft")
            uy_fft = self.compute("uy_fft")
            result = self.oper.rotfft_from_vecfft(ux_fft, uy_fft)
        elif key == "q":
            rot = self.get_var("rot")
            result = rot
        elif key == "div_fft":
            ux_fft = self.compute("ux_fft")
            uy_fft = self.compute("uy_fft")
            result = self.oper.divfft_from_vecfft(ux_fft, uy_fft)
        elif key == "div":
            div_fft = self.compute("div_fft")
            result = self.oper.ifft2(div_fft)
        elif key == "ap_fft":
            rot_fft = self.state_spect.get_var("rot_fft")
            ux_fft, uy_fft = self.oper.vecfft_from_rotfft(rot_fft)
            # print("ux_fft ap fft", ux_fft)
            # uy_fft = self.oper.fft2(self.state_phys.get_var("uy"))
            b_fft = self.state_spect.get_var("b_fft")
            N = self.sim.params.N
            omega_k = self.sim.compute_dispersion_relation()
            result = (N ** 2) * uy_fft + 1j * omega_k * b_fft
            # result = (N ** 2) * uy_fft - 1j * omega_k * b_fft
        elif key == "am_fft":
            rot_fft = self.state_spect.get_var("rot_fft")
            # print("rot_fft", rot_fft)
            ux_fft, uy_fft = self.oper.vecfft_from_rotfft(rot_fft)
            # print("ux_fft am fft", ux_fft)
            # print("uy_fft", uy_fft)
            # uy_fft = self.oper.fft2(self.state_phys.get_var("uy"))
            b_fft = self.state_spect.get_var("b_fft")
            # print("b_fft", b_fft)
            N = self.sim.params.N
            omega_k = self.sim.compute_dispersion_relation()
            # print("omega_k", omega_k)
            # print("1j * omega_k * b_fft", 1j * omega_k * b_fft)
            # print("(N ** 2) * uy_fft", (N ** 2) * uy_fft)
            result = (N ** 2) * uy_fft - 1j * omega_k * b_fft
            # result = (N ** 2) * uy_fft + 1j * omega_k * b_fft
        # result = np.zeros(self.oper.shapeK, dtype="complex")
        elif key == "ap":
            ap_fft = self.compute("ap_fft")
            result = self.oper.ifft(ap_fft)
        elif key == "am":
            am_fft = self.compute("am_fft")
            result = self.oper.ifft(am_fft)

        else:
            to_print = 'Do not know how to compute "' + key + '".'
            if RAISE_ERROR:
                raise ValueError(to_print)

            else:
                if mpi.rank == 0:
                    print(to_print + "\nreturn an array of zeros.")

                result = self.oper.create_arrayX(value=0.0)

        if SAVE_IN_DICT:
            self.vars_computed[key] = result
            self.it_computed[key] = it

        return result

    def statephys_from_statespect(self):
        """Compute `state_phys` from `statespect`."""
        rot_fft = self.state_spect.get_var("rot_fft")
        b_fft = self.state_spect.get_var("b_fft")
        ux_fft, uy_fft = self.oper.vecfft_from_rotfft(rot_fft)

        rot = self.state_phys.get_var("rot")
        ux = self.state_phys.get_var("ux")
        uy = self.state_phys.get_var("uy")
        b = self.state_phys.get_var("b")

        self.oper.ifft_as_arg(rot_fft, rot)
        self.oper.ifft_as_arg(ux_fft, ux)
        self.oper.ifft_as_arg(uy_fft, uy)
        self.oper.ifft_as_arg(b_fft, b)

    def statespect_from_statephys(self):
        """Compute `state_spect` from `state_phys`."""

        rot = self.state_phys.get_var("rot")
        b = self.state_phys.get_var("b")

        rot_fft = self.state_spect.get_var("rot_fft")
        b_fft = self.state_spect.get_var("b_fft")

        self.oper.fft_as_arg(rot, rot_fft)
        self.oper.fft_as_arg(b, b_fft)

    def init_from_rotbfft(self, rot_fft, b_fft):
        """Initialize the state from the variable rot_fft and b_fft."""
        self.oper.dealiasing(rot_fft)
        self.oper.dealiasing(b_fft)

        self.state_spect.set_var("rot_fft", rot_fft)
        self.state_spect.set_var("b_fft", b_fft)

        self.statephys_from_statespect()

    def init_from_apfft(self, ap_fft):
        """Initialize the state from the linear mode ap_fft. """
        rot_fft, b_fft = self.compute_rotbfft_from_apfft(ap_fft)
        self.init_from_rotbfft(rot_fft, b_fft)

    def init_from_amfft(self, am_fft):
        """Initialize the state from the linear mode am_fft. """
        rot_fft, b_fft = self.compute_rotbfft_from_amfft(am_fft)
        self.init_from_rotbfft(rot_fft, b_fft)

    def compute_rotbfft_from_amfft(self, am_fft):
        """
        Computes rot_fft and b_fft from linear mode am_fft.

        ap_fft is assumed to be null
        """
        # it should be (ap_fft + am_fft)
        # but ap_fft is assumed to be null
        uy_fft = (1.0 / (2 * self.params.N ** 2)) * am_fft
        cond = self.oper.KX != 0
        division = np.zeros_like(self.oper.KY)
        division[cond] = self.oper.KY[cond] / self.oper.KX[cond]
        ux_fft = -1 * division * uy_fft

        rot_fft = self.oper.rotfft_from_vecfft(ux_fft, uy_fft)

        omega_k = self.params.N * self.oper.KX / self.oper.K_not0
        b_fft = np.zeros(self.oper.shapeK, dtype="complex")

        # it should be (ap_fft[cond] - am_fft[cond])
        # but ap_fft is assumed to be null
        b_fft[cond] = (1.0 / (2j * omega_k[cond])) * (-am_fft[cond])
        return rot_fft, b_fft

    def compute_rotbfft_from_apfft(self, ap_fft):
        """
        Computes rot_fft and b_fft from linear mode ap_fft.

        am_fft is assumed to be null
        """
        # it should be (ap_fft + am_fft)
        # but am_fft is assumed to be null
        uy_fft = (1.0 / (2 * self.params.N ** 2)) * ap_fft
        cond = self.oper.KX != 0
        division = np.zeros_like(self.oper.KY)
        division[cond] = self.oper.KY[cond] / self.oper.KX[cond]
        ux_fft = -1 * division * uy_fft

        rot_fft = self.oper.rotfft_from_vecfft(ux_fft, uy_fft)

        omega_k = self.params.N * self.oper.KX / self.oper.K_not0
        b_fft = np.zeros(self.oper.shapeK, dtype="complex")
        # it should be (ap_fft[cond] - am_fft[cond])
        # but am_fft is assumed to be null
        b_fft[cond] = (1.0 / (2j * omega_k[cond])) * ap_fft[cond]

        return rot_fft, b_fft

    def init_from_rotfft(self, rot_fft):
        """Initialize the state from the variable `rot_fft`."""
        b_fft = self.oper.create_arrayK(value=0.0)
        self.init_from_rotbfft(rot_fft, b_fft)

    def init_statespect_from(self, **kwargs):
        if len(kwargs) == 1:
            key, arr = kwargs.popitem()
            if key == "rot_fft":
                self.init_from_rotfft(arr)
            elif key == "ap_fft":
                self.init_from_apfft(arr)
            elif key == "am_fft":
                self.init_from_amfft(arr)
            elif key == "ux_fft":
                self.init_from_uxfft(arr)
            elif key == "uy_fft":
                self.init_from_uyfft(arr)
            else:
                super(StateNS2D, self).init_statespect_from(**kwargs)
        elif len(kwargs) == 2:
            if "rot_fft" in kwargs and "b_fft" in kwargs:
                self.init_from_rotbfft(kwargs["rot_fft"], kwargs["b_fft"])
            else:
                super(StateNS2D, self).init_statespect_from(**kwargs)
        else:
            super(StateNS2D, self).init_statespect_from(**kwargs)



import unittest

import fluiddyn.util.mpi as mpi
from fluidsim.util.testing import (
    TestSimulConserveOutput,
    classproperty,
    skip_if_no_fluidfft,
)


@skip_if_no_fluidfft
class TestSimulSW1LExactlin(TestSimulConserveOutput):
    zero = 1e-3

    @classproperty
    def Simul(cls):
        from fluidsim.solvers.sw1l.exactlin.solver import Simul

        return Simul

    @classmethod
    def init_params(cls):
        super().init_params()
        params = cls.params
        params.output.HAS_TO_SAVE = True

        params.output.periods_save.spect_energy_budg = 0.2

        params.forcing.enable = True
        params.forcing.type = "waves_vortices"
        params.forcing.forcing_rate = 0.01
        params.forcing.nkmin_forcing = 2
        params.forcing.nkmax_forcing = 4
        params.init_fields.type = "noise"
        params.oper.nx = 16
        params.oper.ny = 8

    def get_tendencies(self):
        tendencies_fft = self.tendencies_fft

        Fq_fft = tendencies_fft.get_var("q_fft")
        Fap_fft = tendencies_fft.get_var("ap_fft")
        Fam_fft = tendencies_fft.get_var("am_fft")

        return self.sim.oper.uxuyetafft_from_qapamfft(Fq_fft, Fap_fft, Fam_fft)

    def test_enstrophy_conservation(self):
        """Verify that the enstrophy growth rate due to nonlinear tendencies
        (advection term) must be zero.

        .. FIXME: Improve this test

        """
        tendencies_fft = self.tendencies_fft
        return
        state_spect = self.sim.state.state_spect
        oper = self.sim.oper
        Fq_fft = tendencies_fft.get_var("q_fft")
        q_fft = state_spect.get_var("q_fft")

        T_q = (Fq_fft.conj() * q_fft + Fq_fft * q_fft.conj()).real / 2.0
        sum_T = oper.sum_wavenumbers(T_q)
        self.assertAlmostZero(sum_T, tolerance_warning=False)

    def test_energy_conservation(self):
        """Verify that the energy growth rate due to nonlinear tendencies
        (advection term) must be zero.

        """
        Fx_fft, Fy_fft, Feta_fft = self.get_tendencies()
        state_phys = self.sim.state.state_phys
        ux = state_phys.get_var("ux")
        uy = state_phys.get_var("uy")
        eta = state_phys.get_var("eta")

        oper = self.sim.oper
        Fx = oper.ifft2(Fx_fft)
        Fy = oper.ifft2(Fy_fft)
        Feta = oper.ifft2(Feta_fft)
        A = (
            Feta * (ux ** 2 + uy ** 2) / 2
            + (1 + eta) * (ux * Fx + uy * Fy)
            + self.sim.params.c2 * eta * Feta
        )

        A_fft = oper.fft2(A)
        if mpi.rank == 0:
            self.assertAlmostZero(A_fft[0, 0], tolerance_warning=False)


if __name__ == "__main__":
    unittest.main()



import numpy as np

from fluiddyn.util import mpi

from .base import SpecificOutput


def cumsum_inv(a):
    return a[::-1].cumsum()[::-1]


def inner_prod(a_fft, b_fft):
    return np.real(a_fft.conj() * b_fft)


class SpectralEnergyBudgetBase(SpecificOutput):
    """Handle the saving and plotting of spectral energy budget.

    This class uses the particular functions defined by some solvers
    :func:`` and
    :func``. If the solver doesn't has these
    functions, this class does nothing.
    """

    _tag = "spect_energy_budg"
    _name_file = _tag + ".h5"

    @staticmethod
    def _complete_params_with_default(params):
        tag = "spect_energy_budg"

        params.output.periods_save._set_attrib(tag, 0)
        params.output._set_child(tag, attribs={"HAS_TO_PLOT_SAVED": False})

    def __init__(self, output):

        params = output.sim.params
        self.nx = params.oper.nx

        self.spectrum2D_from_fft = output.sim.oper.spectrum2D_from_fft
        self.spectra1D_from_fft = output.sim.oper.spectra1D_from_fft
        self.sum_wavenumbers = output.sim.oper.sum_wavenumbers

        HAS_TO_PLOT_SAVED = params.output.spect_energy_budg.HAS_TO_PLOT_SAVED
        super().__init__(
            output,
            period_save=params.output.periods_save.spect_energy_budg,
            has_to_plot_saved=HAS_TO_PLOT_SAVED,
            arrays_1st_time={
                "khE": output.sim.oper.khE,
                "kxE": output.sim.oper.kxE,
                "kyE": output.sim.oper.kyE,
            },
        )

    def compute(self):
        """compute the values at one time."""
        if mpi.rank == 0:
            dict_results = {}
            return dict_results

    def _init_online_plot(self):
        if mpi.rank == 0:
            width_axe = 0.85
            height_axe = 0.37
            x_left_axe = 0.12
            z_bottom_axe = 0.56

            size_axe = [x_left_axe, z_bottom_axe, width_axe, height_axe]
            self.fig, axe_a = self.output.figure_axe(
                size_axe=size_axe, numfig=4_000_000
            )
            self.axe_a = axe_a
            axe_a.set_xlabel(r"$k_h$")
            axe_a.set_ylabel(r"$\Pi(k_h)$")
            axe_a.set_title("energy flux\n" + self.output.summary_simul)
            axe_a.set_xscale("log")

            z_bottom_axe = 0.08
            size_axe[1] = z_bottom_axe
            axe_b = self.fig.add_axes(size_axe)
            self.axe_b = axe_b
            axe_b.set_xlabel(r"$k_h$")
            axe_b.set_ylabel(r"$\Pi(k_h)$")
            axe_b.set_xscale("log")

    def fnonlinfft_from_uxuy_funcfft(self, ux, uy, f_fft):
        r"""
        Compute a non-linear term.

        Notes
        -----
        Returns an fft-sized nd-array equivalent to the expression:

        .. math:: - \widehat{(\vec{u}.\nabla)f}
        """

        oper = self.oper
        px_f_fft, py_f_fft = oper.gradfft_from_fft(f_fft)
        px_f = oper.ifft2(px_f_fft)
        py_f = oper.ifft2(py_f_fft)
        del (px_f_fft, py_f_fft)
        Fnl = -ux * px_f - uy * py_f
        del (px_f, py_f)
        Fnl_fft = oper.fft2(Fnl)
        oper.dealiasing(Fnl_fft)
        return Fnl_fft

    def fnonlinfft_from_uruddivfunc(
        self, urx, ury, udx, udy, div, func_fft, func
    ):
        """Compute a non-linear term."""
        oper = self.oper
        px_func_fft, py_func_fft = oper.gradfft_from_fft(func_fft)
        px_func = oper.ifft2(px_func_fft)
        py_func = oper.ifft2(py_func_fft)
        del (px_func_fft, py_func_fft)
        Frf = -urx * px_func - ury * py_func
        Fdf = -udx * px_func - udy * py_func - div * func / 2
        del (px_func, py_func)
        Frf_fft = oper.fft2(Frf)
        Fdf_fft = oper.fft2(Fdf)
        oper.dealiasing(Frf_fft, Fdf_fft)
        return Frf_fft, Fdf_fft



"""Mini operator to modify the resolution (Fourier)
===================================================

Internal API
------------

.. autofunction:: fill_field_fft_3d

.. autoclass:: MiniOperModifResol
   :members:
   :private-members:

"""

import numpy as np

from fluiddyn.calcul.easypyfft import FFTW2DReal2Complex, FFTW3DReal2Complex

from fluidsim.base.init_fields import fill_field_fft_2d

from transonic import boost, Array

A = Array[np.complex128, "3d", "C"]


@boost
def fill_field_fft_3d(field_fft_in: A, field_fft_out: A):
    """Fill the values from field_fft_in in field_fft_out

    This function is specialized for FFTW3DReal2Complex (no MPI).
    """

    [nk0_out, nk1_out, nk2_out] = field_fft_out.shape
    [nk0_in, nk1_in, nk2_in] = field_fft_in.shape

    nk0_min = min(nk0_out, nk0_in)
    nk1_min = min(nk1_out, nk1_in)
    nk2_min = min(nk2_out, nk2_in)

    for ik0 in range(nk0_min // 2 + 1):
        for ik1 in range(nk1_min // 2 + 1):
            for ik2 in range(nk2_min):
                # positive wavenumbers
                field_fft_out[ik0, ik1, ik2] = field_fft_in[ik0, ik1, ik2]
                # negative wavenumbers
                if ik0 > 0 and ik0 < nk0_min // 2:
                    field_fft_out[-ik0, ik1, ik2] = field_fft_in[-ik0, ik1, ik2]
                    if ik1 > 0 and ik1 < nk1_min // 2:
                        field_fft_out[-ik0, -ik1, ik2] = field_fft_in[
                            -ik0, -ik1, ik2
                        ]
                if ik1 > 0 and ik1 < nk1_min // 2:
                    field_fft_out[ik0, -ik1, ik2] = field_fft_in[ik0, -ik1, ik2]


class MiniOperModifResol:
    def __init__(self, shape):

        self.shape = shape

        dimension = self.dimension = len(shape)

        if dimension not in [2, 3]:
            raise NotImplementedError

        if dimension == 2:
            ny, nx = shape
            self.oper_fft = FFTW2DReal2Complex(nx, ny)
            self.axes = tuple("xy")
        else:
            nz, ny, nx = shape
            self.oper_fft = FFTW3DReal2Complex(nx, ny, nz)
            self.axes = tuple("xyz")

        self.fft_as_arg = self.oper_fft.fft_as_arg
        self.ifft_as_arg = self.oper_fft.ifft_as_arg

        self.create_arrayX = self.oper_fft.create_arrayX
        self.create_arrayK = self.oper_fft.create_arrayK

    def fill_field_fft(self, field_spect, field2_spect, oper):

        if self.dimension == 2:
            return fill_field_fft_2d(field_spect, field2_spect)

        fill_field_fft_3d(field_spect, field2_spect)
